################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: index2gff.py 2880 2010-04-07 08:44:13Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
index2bed.py - convert indexed fasta file to bed file
=====================================================

:Author: Andreas Heger
:Release: $Id: index2gff.py 2880 2010-04-07 08:44:13Z andreas $
:Date: |today|
:Tags: Python

Purpose
-------

Usage
-----

Type::

   python <script_name>.py --help

for command line help.

Code
----

""" 

import os
import sys
import re
import optparse

import CGAT.IndexedFasta as IndexedFasta
import CGAT.Bed as Bed
import CGAT.Experiment as E

def getFixedWidthWindows( map_contig2size, window_size ):
    """return a list of fixed contig sizes."""
    

    for contig, size in map_contig2size.items():
        E.info("processing %s" % contig)
        for x in range(0, size, window_increment):
            if x + window_size > size: continue
            gff = GTF.Entry()
            gff.feature = "window"
            gff.source = "window"
            gff.contig = contig
            gff.start = x
            gff.end = min(size, x + window_size)
            yield gff


def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: index2gff.py 2880 2010-04-07 08:44:13Z andreas $",
                                    usage = globals()["__doc__"] )

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default]."  )

    parser.add_option( "--remove-regex", dest="remove_regex", type="string", 
                       help="regular expression of contigs to remove [default=None]."  )

    parser.add_option( "-e", "--gff-file", dest="gff_file", type="string",
                       help="gff file to use for getting contig sizes."  )

    parser.add_option( "-f", "--fixed-width-windows=", dest="fixed_width_windows", type="string",
                       help="fixed width windows. Supply the window size as a parameter. Optionally supply an offset." )

    parser.set_defaults(
        genome_file = None,
        remove_regex = None,
        fixed_windows = None,
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    if options.remove_regex:
        remove_regex = re.compile( options.remove_regex )
    else:
        remove_regex = None

    if options.fixed_width_windows:
        v = map(int, options.fixed_width_windows.split(","))
        if len(v) == 2:
            window_size, window_increment = v
        elif len(v) == 1:
            window_size, window_increment = v[0], v[0]
        else:
            raise ValueError( "could not parse window size '%s': should be size[,increment]" % options.fixed_width_windows )

    if options.gff_file:
        infile = open( options.gff_file, "r" )
        gff = GTF.readFromFile( infile )   
        infile.close()
        for g in gff:
            try:
                map_contig2size[g.mName] = max( map_contig2size[g.mName], g.end )
            except ValueError:
                map_contig2size[g.mName] = g.end 
            
    else:
        gff = None

    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta( options.genome_file )
        map_contig2size = fasta.getContigSizes( with_synonyms = False )
    else:
        fasta = None

    if map_contig2size == None:
        raise ValueError( "no source of contig sizes supplied" )

    ## do sth
    counter = E.Counter()

    for contig, size in map_contig2size.items():
        size = int(size)
        counter.input += 1

        if remove_regex and remove_regex.search(contig): 
            counter.skipped += 1
            continue

        if options.fixed_width_windows:
            for x in range(0, size, window_increment):
                if x + window_size > size: continue
                options.stdout.write( "%s\t%i\t%i" % (contig, x, min(size, x + window_size)))
                counter.windows += 1
        else:
            options.stdout.write( "%s\t%i\t%i" % (contig, 0, size))
            counter.windows += 1

        counter.output += 1

    E.info( str(counter) )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )




from cnestedlist import *

import cPickle, sqlite3
import os

sqlite3.register_converter("pickle", cPickle.loads)

class NCLSimple(object):
    """a nested contained list in memory storing
    no additional data.

    If *filestem* is not given, an new database is created
    in memory. If *filestem* is given, a database is opened 
    from disk. If it does not exist or *force* is set to True, 
    a new database is created. 

    Convenience wrapper around low-level ncl functions.

    Note that NCL objects are most efficient
    if used in write-once/read many applications.
    """
    
    def __init__(self, filestem = None, force = False):
        self.mTuples = []
        self.mIsDirty = False
        if filestem != None:
            self.mFilestem = filestem
            if not force and os.path.exists( os.path.abspath(filestem) + ".idb" ):
                self.mFromDisk = True
                self.mDatabase = IntervalFileDB( filestem )
            else:
                self.mFromDisk = False
                self.mDatabase = IntervalDB()
        else:
            self.mFilestem = None
            self.mFromDisk = False
            self.mDatabase = IntervalDB()

    def add( self, start, end ):
        """add segment *start*,*end* with value to database.

        returns the index of the added segment.
        """
        assert self.mFromDisk == False, "can not add to pre-existing or flushed databases"
        if start < 0: raise ValueError( "only positive coordinates are accepted (%i<0)" % start )
        if start >= end:
            raise ValueError( "adding empty/invalid interval (%i,%i)" % (start,end) )
        v = len(self.mTuples)
        self.mTuples.append( (start, end, v) )
        self.mIsDirty = True
        return v

    def find( self, start, end ):
        """find intervals in database overlapping with *start* and *end*.
    
        returns an :class:`ncl.IntervalDBIterator`
        """
        if start < 0: raise ValueError( "only positive coordinates are accepted (%i<0)" % start )
        self._commit()
        return self.mDatabase.find_overlap( start, end )

    def _commit( self ):
        """commit database if changed."""
        if self.mIsDirty:
            self.mDatabase.fromlist( self.mTuples )
            self.mIsDirty = False
    def __del__(self):
        """flush database to disk."""
        if self.mFilestem and not self.mFromDisk:
            if self.mIsDirty:
                self.mDatabase.fromlist( self.mTuples )
            # flush database
            self.mDatabase.write_binaries( self.mFilestem )

class NCL(NCLSimple):
    """a nested contained list in memory. This
    class stores information on intervals.
    
    Note that NCL objects are most efficient
    if used in write-once/read many applications.
    """
    
    def __init__(self, *args, **kwargs):
        NCLSimple.__init__(self,*args,**kwargs)
        self.mValues = []
        # route calls to __getitem__ directly to list if in memory
        if self.mFromDisk:
            fn = self.mFilestem + ".vals" 
            self.mDBHandle = sqlite3.connect( fn )

    def add( self, start, end, value ):
        """add segment *start*,*end* with value to database.

        returns the index of the added segment.
        """
        assert start >= 0, "only positive coordinates are accepted (%i<0)" % start
        if start >= end:
            raise ValueError( "adding empty/invalid interval (%i,%i)" % (start,end) )
        self.mValues.append( value )
        return NCLSimple.add( self, start, end )

    def __getitem__(self, key):
        """get a value from the database
        """
        cc = self.mDBHandle.cursor()
        val = cc.execute( "SELECT value FROM data WHERE id = '%i'" % key ).fetchone()[0]
        cc.close()
        return cPickle.loads( str(val) )

    def find( self, start, end ):
        """find intervals overlapping *start* and *end*.

        returns an :class:`ncl.IteratorWithValues`
        """
        if self.mFromDisk:
            return IteratorWithValues( self, NCLSimple.find( self, start, end ))
        else:
            return IteratorWithValues( self.mValues, NCLSimple.find( self, start, end ))

    def _commit( self ):
        """commit database if changed."""
        if self.mIsDirty and self.mFilestem:
            # flush database
            self._flushValues()

        NCLSimple._commit(self)

    def _flushValues( self ):
        """flush values to disk."""

        fn = self.mFilestem + ".vals" 
        if os.path.exists( fn ): os.remove( fn )
        dbhandle = sqlite3.connect( fn )

        cc = dbhandle.cursor()
        cc.execute( """create table data( id INTEGER, value BLOB);""" )
        cc.executemany( """INSERT INTO data VALUES (?,?)""", list(enumerate( map( lambda x: sqlite3.Binary(cPickle.dumps(x)), self.mValues )) ))
        dbhandle.commit()
        cc.close()

    def __del__(self):
        """flush database to disk."""
        if self.mFilestem and self.mIsDirty: self._flushValues()

        NCLSimple.__del__( self )

class IteratorWithValues(object):
    """an iterator over intervals with values.
    """
    
    def __init__(self, values, iter ):

        self.mValues = values
        self.mIterator = iter

    def __iter__(self): return self
        
    def next(self):
        start, end, idx = self.mIterator.next()
        return( (start, end, self.mValues[idx] ) )
    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Stats.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''
import types
import math
import numpy
import scipy
import scipy.stats
import scipy.interpolate
import collections
from rpy2.robjects import r as R
import rpy2.robjects as ro
import rpy2.robjects.numpy2ri
rpy2.robjects.numpy2ri.activate()

def getSignificance( pvalue, thresholds=[0.05, 0.01, 0.001] ):
    """return cartoon of significance of a p-Value."""
    n = 0
    for x in thresholds:
        if pvalue > x: return "*" * n
        n += 1
    return "*" * n

class Result(object):
    '''allow both member and dictionary access.'''
    slots=("_data")
    def __init__(self):
        object.__setattr__(self, "_data", dict())
    def fromR( self, take, r_result ):
        '''convert from an *r_result* dictionary using map *take*.

        *take* is a list of tuples mapping a field to the corresponding
        field in *r_result*.
        '''
        for x,y in take:
            if y:
                self._data[x] = r_result.rx(y)[0][0]
            else:
                self._data[x] = r_result.rx(x)[0][0]

            # if y:
            #     self._data[x] = r_result[y]
            # else:
            #     self._data[x] = r_result[x]

        return self
    def __getattr__(self, key):
        if not key.startswith("_"):
            try: return object.__getattribute__(self,"_data")[key]
            except KeyError: pass
        return getattr( self._data, key )
    def keys(self): return self._data.keys()
    def values(self): return self._data.values()
    def __iter__(self): return self._data.__iter__()
    def __len__(self): return self._data.__len__()
    def __str__(self):
        return str(self._data)
    def __contains__(self,key):
        return key in self._data
    def __getitem__(self, key ):
        return self._data[key]
    def __delitem__(self, key ):
        del self._data[key]
    def __setitem__(self, key, value ):
        self._data[key] = value
    def __setattr__(self, key, value):
        if not key.startswith("_"):
            self._data[key] = value
        else:
            object.__setattr__(self,key,value)
    def __getstate__(self):
        # required for correct pickling/unpickling
        return object.__getattribute__(self,"_data")
            
    def __setstate__(self,d):
        # required for correct unpickling, otherwise
        # maximum recursion threshold will be reached
        object.__setattr__(self, "_data", d)

#################################################################
#################################################################
#################################################################
## Perform log likelihood test
class LogLikelihoodTest:

    def __init__(self):
        pass

def doLogLikelihoodTest( complex_ll, complex_np,
                         simple_ll, simple_np,
                         significance_threshold = 0.05):
    """perform log-likelihood test between model1 and model2.
    """

    assert complex_ll >= simple_ll, "log likelihood of complex model smaller than for simple model: %f > %f" % (complex_ll, simple_ll)

    chi = 2 * (complex_ll - simple_ll)
    df = complex_np - simple_np

    if df <= 0:
        raise ValueError, "difference of degrees of freedom not larger than 0"
    
    p = scipy.stats.chisqprob( chi, df )
    
    l = LogLikelihoodTest()
    
    l.mComplexLogLikelihood = complex_ll
    l.mSimpleLogLikelihood = simple_ll    
    l.mComplexNumParameters = complex_np
    l.mSimpleNumParameters = simple_np    
    l.mSignificanceThreshold = significance_threshold
    l.mProbability = p
    l.mChiSquaredValue = chi
    l.mDegreesFreedom = df

    if p < significance_threshold:
        l.mPassed = True
    else:
        l.mPassed = False
        
    return l

#################################################################
#################################################################
#################################################################
class BinomialTest:
    def __init__(self):
        pass

def doBinomialTest( p, sample_size, observed, significance_threshold = 0.05):
    """perform a binomial test. 
    
    Given are p: the probability of the NULL hypothesis, the sample_size
    and the number of observed counts.
    """
    pass

#################################################################
#################################################################
#################################################################
class ChiSquaredTest:
    def __init__(self):
        pass

def doChiSquaredTest( matrix, significance_threshold = 0.05 ):
    '''perform chi-squared test on a matrix.

    The observed/expected values are in rows, the categories are in columns, for
    example:

    +---------+--------------+--------+----------+
    |set      |protein_coding|intronic|intergenic|
    +---------+--------------+--------+----------+
    |observed |92            |90      |194       |
    +---------+--------------+--------+----------+
    |expected |91            |10      |15        |
    +---------+--------------+--------+----------+

    If there are only two categories (one degrees of freedom) the Yates correction is applied.
    For each entry (observed-expected), the value 0.5 is subtracted ignoring the sign of the difference.

    The test throws an exception if
    1. one or more expected categories are less than 1 (it does not matter what the observed values are)
    2. more than one-fifth of expected categories are less than 5
    '''

    nrows, ncols = matrix.shape

    if nrows != 2:
        raise NotImplementedError( "chi-square currently only implemented for 2xn tables." )
    
    n = 0
    for x in range(ncols):
        if matrix[1][x] < 1: 
            raise ValueError( "matrix contains expected counts < 1" )
        if matrix[1][x] < 5: n +=1
    if 100.0 * n / ncols > 20.0:
        raise ValueError( "more than 20% of expected categories are less than 5" )

    row_sums = [ sum(matrix[x,:]) for x in range( nrows ) ]
    col_sums = [ sum(matrix[:,x]) for x in range( ncols ) ]
    sample_size = float(sum(row_sums))
    chi = 0.0

    df = (nrows - 1) * (ncols -1 )
    ## Yates correction applies for a 2x2 table only (df==1)
    if df == 1: 
        correction = 0.5 * 0.5
    else:
        correction = 0
        
    for x in range(nrows):
        for y in range(ncols):
            expected = row_sums[x] * col_sums[y] / sample_size
            # compute difference and apply Yates correction
            d = abs(matrix[x,y] - expected) - correction
            chi += (d * d ) / expected

    result = ChiSquaredTest()

    result.mProbability = scipy.stats.chisqprob( chi, df )
    result.mDegreesFreedom = df
    result.mChiSquaredValue = chi
    result.mPassed = result.mProbability < significance_threshold
    result.mSignificance = getSignificance( result.mProbability )
    result.mSampleSize = sample_size
    result.mPhi = math.sqrt( result.mChiSquaredValue / result.mSampleSize )
    return result

def doPearsonChiSquaredTest( p, sample_size, observed, significance_threshold = 0.05):
    """perform a pearson chi squared test.
    
    Given are p: the probability of the NULL hypothesis, the sample_size
    and the number of observed counts.
    
    For large sample sizes, this test is a continuous approximation to
    the binomial test.
    """
    e = float(p) * sample_size
    d = float(observed) - e
    chi = d * d / e
    df = 1

    result = ChiSquaredTest()

    result.mProbability = scipy.stats.chisqprob( chi, df )
    result.mDegreesFreedom = df
    result.mChiSquaredValue = chi
    result.mPassed = result.mProbability < significance_threshold
    result.mSignificance = getSignificance( result.mProbability )
    result.mSampleSize = sample_size
    result.mPhi = math.sqrt( result.mChiSquaredValue / result.mSampleSize )
    result.mObserved = observed
    result.mExpected = e
    return result

#################################################################
#################################################################
#################################################################
## Convenience functions and objects for statistical analysis

class DistributionalParameters:
    """a collection of distributional parameters. Available properties
    are:

    mMean, mMedian, mMin, mMax, mSampleStd, mSum, mCounts

    This method is deprecated - use :class:`Summary` instead.
    """
    def __init__(self, values = None, format = "%6.4f", mode="float"):

        self.mMean, self.mMedian, self.mMin, self.mMax, self.mSampleStd, self.mSum, self.mCounts, self.mQ1, self.mQ3 = \
                    (0, 0, 0, 0, 0, 0, 0, 0, 0)
        
        if values != None and len(values) > 0: self.updateProperties( values )
        self.mFormat = format
        self.mMode = mode
        self.mNErrors = 0

    def updateProperties( self, values):
        """update properties.

        If values is an vector of strings, each entry will be converted
        to float. Entries that can not be converted are ignored.
        """
        values = [x for x in values if x != None ]

        if len(values) == 0:
            raise ValueError( "no data for statistics" )

        ## convert
        self.mNErrors = 0
        if type(values[0]) not in (types.IntType, types.FloatType):
            n = []
            for x in values:
                try:
                    n.append( float(x) )
                except ValueError:
                    self.mNErrors += 1
        else:
            n = values

        if len(n) == 0:
            raise ValueError( "no data for statistics" )

        ## use a non-sort algorithm later.
        n.sort()
        self.mQ1 = n[len(n) / 4]
        self.mQ3 = n[len(n) * 3 / 4]

        self.mCounts = len(n)
        self.mMin = min(n)
        self.mMax = max(n)
        self.mMean = scipy.mean( n )
        self.mMedian = scipy.median( n )
        self.mSampleStd = scipy.std( n )
        self.mSum = reduce( lambda x, y: x+y, n )

    def getZScore( self, value ):
        """return zscore for value."""
        if self.mSampleStd > 0:
            return (value - self.mMean) / self.mSampleStd
        else:
            return 0

    def setFormat( self, format ):
        """set number format."""
        self.mFormat = format

    def getHeaders( self ):
        """returns header of column separated values."""
        return ("nval", "min", "max", "mean", "median", "stddev", "sum", "q1", "q3")

    def getHeader( self ):
        """returns header of column separated values."""
        return "\t".join( self.getHeaders())

    def items(self):
        return [ (x, self.__getitem__(x)) for x in self.getHeaders() ]

    def __getitem__( self, key ):
        
        if key == "nval": return self.mCounts
        if key == "min": return self.mMin
        if key == "max": return self.mMax
        if key == "mean": return self.mMean
        if key == "median": return self.mMedian
        if key == "stddev": return self.mSampleStd
        if key == "sum": return self.mSum
        if key == "q1": return self.mQ1
        if key == "q3": return self.mQ3                

        raise KeyError, key
        
    def __str__( self ):
        """return string representation of data."""
        
        if self.mMode == "int":
            format_vals = "%i"
            format_median = "%.1f"
        else:
            format_vals = self.mFormat
            format_median = self.mFormat

        return "\t".join( ( "%i" % self.mCounts,
                            format_vals % self.mMin,
                            format_vals % self.mMax,
                            self.mFormat % self.mMean,
                            format_median % self.mMedian,
                            self.mFormat % self.mSampleStd,                                      
                            format_vals % self.mSum,
                            format_vals % self.mQ1,
                            format_vals % self.mQ3,                            
                            ) )

class Summary( Result ):
    """a collection of distributional parameters. Available properties
    are:

    mean, median, min, max, samplestd, sum, counts
    """

    fields = ("nval", "min", "max", "mean", "median", "stddev", "sum", "q1", "q3")

    def __init__(self, values = None, 
                 format = "%6.4f", mode="float",
                 allow_empty = True ):

        Result.__init__(self)
        self._format = format
        self._mode = mode

        # note that this determintes the order of the fields at output
        self.counts, self.min, self.max, self.mean, self.median, self.samplestd, self.sum, self.q1, self.q3 = \
                    (0, 0, 0, 0, 0, 0, 0, 0, 0)
        
        if values != None:

            values = [x for x in values if x != None ]

            if len(values) == 0:
                if allow_empty: return
                else: raise ValueError( "no data for statistics" )

            # convert
            self._nerrors = 0
            if type(values[0]) not in (types.IntType, types.FloatType):
                n = []
                for x in values:
                    try:
                        n.append( float(x) )
                    except ValueError:
                        self._nerrors += 1
            else:
                n = values

            ## use a non-sort algorithm?
            n.sort()
            if len(n):
                self.q1 = n[len(n) / 4]
                self.q3 = n[len(n) * 3 / 4]
            else:
                self.q1 = self.q3 = 0 

            self.counts = len(n)
            self.min = min(n)
            self.max = max(n)
            self.mean = scipy.mean( n )
            self.median = scipy.median( n )
            self.samplestd = scipy.std( n )
            self.sum = reduce( lambda x, y: x+y, n )

    def getHeaders( self ):
        """returns header of column separated values."""
        return self.fields

    def getHeader( self ):
        """returns header of column separated values."""
        return "\t".join( self.getHeaders())

    def __str__( self ):
        """return string representation of data."""
        
        if self._mode == "int":
            format_vals = "%i"
            format_median = "%.1f"
        else:
            format_vals = self._format
            format_median = self._format

        return "\t".join( ( "%i" % self.counts,
                            format_vals % self.min,
                            format_vals % self.max,
                            self._format % self.mean,
                            format_median % self.median,
                            self._format % self.samplestd,                                      
                            format_vals % self.sum,
                            format_vals % self.q1,
                            format_vals % self.q3,                            
                            ) )

def adjustPValues( pvalues, method ):
    '''adjust P-Values for multiple testing using
    the p.adjust() method in R.

    Possible values of method are:

    c("holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr", "none")
    '''
    return R.p_adjust( pvalues, method )

def smoothPValues( pvalues,
                   vlambda=numpy.arange(0,0.95,0.05),
                   smooth_df = 3,
                   smooth_log_pi0 = False):

    if min(pvalues) < 0 or max(pvalues) > 1:
        raise ValueError( "p-values out of range" )

    if len(vlambda) > 1 and len(vlambda) < 4:
        raise ValueError(" If length of vlambda greater than 1, you need at least 4 values." )

    if len(vlambda) > 1 and (min(vlambda) < 0 or max(vlambda) >= 1):
        raise ValueError( "vlambda must be within [0, 1).")

    m = len(pvalues)

    pi0 = numpy.zeros( len(vlambda), numpy.float )

    for i in range( len(vlambda) ):
        pi0[i] = numpy.mean( [x >= vlambda[i] for x in pvalues ]) / (1.0 -vlambda[i] )

    R.assign( "pi0", pi0)
    R.assign( "vlambda", vlambda)
    print "pi0=", pi0
    
    if smooth_log_pi0:
        pi0 = math.log(pi0)

    R.assign( "smooth_df", smooth_df)

    spi0 = R("""spi0 <- smooth.spline(vlambda,pi0, df = smooth_df)""")
    pi0 = R("""pi0 <- predict( spi0, x = max(vlambda) )$y""")

    print spi0
    if smooth_log_pi0:
        pi0 = math.exp(pi0)

    return pi0

def getPi0( pvalues, 
            vlambda=numpy.arange(0,0.95,0.05), 
            pi0_method="smoother",
            smooth_df = 3,
            smooth_log_pi0 = False):
    '''used within nubiscan.'''
    
    if min(pvalues) < 0 or max(pvalues) > 1:
        raise ValueError( "p-values out of range" )

    if len(vlambda) > 1 and len(vlambda) < 4:
        raise ValueError(" If length of vlambda greater than 1, you need at least 4 values." )

    if len(vlambda) > 1 and (min(vlambda) < 0 or max(vlambda) >= 1):
        raise ValueError( "vlambda must be within [0, 1).")

    m = len(pvalues)

     # these next few functions are the various ways to estimate pi0
    if len(vlambda)==1: 
        vlambda = vlambda[0]
        if  vlambda < 0 or vlambda >=1 :
            raise ValueError( "vlambda must be within [0, 1).")

        pi0 = numpy.mean( [ x >= vlambda for x in pvalues ] ) / (1.0 - vlambda)
        pi0 = min(pi0, 1.0)
        R.assign( "pi0", pi0)
        
    else:
        
        pi0 = numpy.zeros( len(vlambda), numpy.float )

        for i in range( len(vlambda) ):
            pi0[i] = numpy.mean( [x >= vlambda[i] for x in pvalues ]) / (1.0 -vlambda[i] )

        R.assign( "pi0", pi0)
        R.assign( "vlambda", vlambda)

        if pi0_method=="smoother":
            if smooth_log_pi0:
                pi0 = math.log(pi0)
                
            R.assign( "smooth_df", smooth_df)

            spi0 = R("""spi0 <- smooth.spline(vlambda,pi0, df = smooth_df)""")
            pi0 = R("""pi0 <- predict( spi0, x = max(vlambda) )$y""")

            if smooth_log_pi0:
                pi0 = math.exp(pi0)

        elif pi0_method=="bootstrap":

            minpi0 = min(pi0)

            mse = numpy.zeros( len(vlambda), numpy.float )
            pi0_boot = numpy.zeros( len(vlambda), numpy.float )

            R.assign( "pvalues", pvalues)
            pi0 = R("""
            m <- length(pvalues)
            minpi0 <- min(pi0)
            mse <- rep(0,length(vlambda))
            pi0_boot <- rep(0,length(vlambda))
            for(i in 1:100) 
            {
                pvalues_boot <- sample(pvalues,size=m,replace=TRUE)
                for(i in 1:length(vlambda)) 
                {
                    pi0_boot[i] <- mean(pvalues_boot>vlambda[i])/(1-vlambda[i])
                }
                mse <- mse + (pi0_boot-minpi0)^2
            }
            pi0 <- min(pi0[mse==min(mse)])""")
        else:
            raise ValueError( "'pi0_method' must be one of 'smoother' or 'bootstrap'.")

        pi0 = min(pi0,1.0)

    if pi0 <= 0:
        raise ValueError( "The estimated pi0 <= 0. Check that you have valid p-values or use another vlambda method." )

    return pi0


class FDRResult:
    def __init__(self):
        pass

    def plot(self, hardcopy = None):

        if hardcopy:
            R.png(hardcopy, width=1024, height=768, type="cairo")

        R.require('qvalue')

        # build a qobj
        R.assign( "pval", self.mPValues )
        R.assign( "pi0", self.mPi0 )
        R.assign( "qval", self.mQValues )
        R.assign( "lambda", self.mLambda )
        R("""qobj <-list( pi0=pi0, qvalues=qval, pvalues=pval, lambda=lambda)""")
        R(""" class(qobj) <- "qvalue" """)

        R("""qplot(qobj)""")

        if hardcopy:
            R.dev_off()

def doFDR(pvalues, 
          vlambda=None,
          pi0_method="smoother", 
          fdr_level=None, 
          robust=False,
          smooth_df = 3,
          smooth_log_pi0 = False,
          plot = False ):
    """modeled after code taken from http://genomics.princeton.edu/storeylab/qvalue/linux.html.

    I did not like the error handling so I translated most to python.
    
    Compute FDR after method by Storey et al. (2002).
    """

    # set to default of qvalue method
    if vlambda == None: vlambda = numpy.arange(0,0.95,0.05)

    if min(pvalues) < 0 or max(pvalues) > 1:
        raise ValueError( "p-values out of range" )

    if type(vlambda) == float:
        vlambda = (vlambda, )

    if len(vlambda) > 1 and len(vlambda) < 4:
        raise ValueError(" If length of vlambda greater than 1, you need at least 4 values." )

    if len(vlambda) > 1 and (min(vlambda) < 0 or max(vlambda) >= 1):
        raise ValueError( "vlambda must be within [0, 1).")

    m = len(pvalues)

     # these next few functions are the various ways to estimate pi0
    if len(vlambda)==1: 
        vlambda = vlambda[0]
        if  vlambda < 0 or vlambda >=1 :
            raise ValueError( "vlambda must be within [0, 1).")

        pi0 = numpy.mean( [ x >= vlambda for x in pvalues ] ) / (1.0 - vlambda)
        pi0 = min(pi0, 1.0)
        R.assign( "pi0", pi0)
    else:
        pi0 = numpy.zeros( len(vlambda), numpy.float )

        for i in range( len(vlambda) ):
            pi0[i] = numpy.mean( [x >= vlambda[i] for x in pvalues ]) / (1.0 -vlambda[i] )

        R.assign( "pi0", pi0)
        R.assign( "vlambda", vlambda)

        if pi0_method=="smoother":
            if smooth_log_pi0:
                pi0 = math.log(pi0)
                
            R.assign( "smooth_df", smooth_df)
            spi0 = R("""spi0 <- smooth.spline(vlambda,pi0, df = smooth_df)""")
            if plot:
                import matplotlib.pyplot as plt
                plt.figure()
                plt.plot( vlambda, pi0 )
                x2 = numpy.arange( 0, 1, 0.001 )
                R.assign( "x2", x2)
                y2 = R("""y2 <- predict( spi0, x = x2 )$y""")
                plt.plot( x2, y2 )
                plt.show()
            
            pi0 = R("""pi0 <- predict( spi0, x = max(vlambda) )$y""")[0]

            if smooth_log_pi0:
                pi0 = math.exp(pi0)

        elif pi0_method=="bootstrap":

            minpi0 = min(pi0)

            mse = numpy.zeros( len(vlambda), numpy.float )
            pi0_boot = numpy.zeros( len(vlambda), numpy.float )

            R.assign( "pvalues", pvalues)
            pi0 = R("""
            m <- length(pvalues)
            minpi0 <- min(pi0)
            mse <- rep(0,length(vlambda))
            pi0_boot <- rep(0,length(vlambda))
            for(i in 1:100) 
            {
                pvalues_boot <- sample(pvalues,size=m,replace=TRUE)
                for(i in 1:length(vlambda)) 
                {
                    pi0_boot[i] <- mean(pvalues_boot>vlambda[i])/(1-vlambda[i])
                }
                mse <- mse + (pi0_boot-minpi0)^2
            }
            pi0 <- min(pi0[mse==min(mse)])""")[0]
        else:
            raise ValueError( "'pi0_method' must be one of 'smoother' or 'bootstrap'.")

        pi0 = min(pi0,1.0)
        R.assign( "pi0", pi0 )

    if pi0 <= 0:
        raise ValueError( "The estimated pi0 (%f) <= 0. Check that you have valid p-values or use another vlambda method." % pi0)

    if fdr_level != None and (fdr_level <= 0 or fdr_level > 1):
        raise ValueError( "'fdr_level' must be within (0, 1].")

    # The estimated q-values calculated here
    #u = numpy.argsort( p )

    # change by Alan
    # ranking function which returns number of observations less than or equal
    ro.globalenv['pvalues'] = ro.FloatVector( pvalues )
    R.assign( "robust", robust )
    qvalues = R("""u <- order(pvalues)
    qvalues.rank <- function(x) 
{
      idx <- sort.list(x)

      fc <- factor(x)
      nl <- length(levels(fc))
      bin <- as.integer(fc)
      tbl <- tabulate(bin)
      cs <- cumsum(tbl)

      tbl <- rep(cs, tbl)
      tbl[idx] <- tbl

      return(tbl)
}

v <- qvalues.rank(pvalues)
m <- length(pvalues)
qvalues <- pi0 * m * pvalues / v

if(robust) 
{
        qvalues <- pi0*m*pvalues/(v*(1-(1-pvalues)^m))
}
qvalues[u[m]] <- min(qvalues[u[m]],1)

rqvalues <- qvalues
for(i in (m-1):1) 
{
   qvalues[u[i]] <- min(qvalues[u[i]],qvalues[u[i+1]],1)
}


qvalues
""")

    result = FDRResult()
    result.mQValues = qvalues

    if fdr_level != None:
        result.mPassed = [ x <= fdr_level for x in result.mQValues ]
    else:
        result.mPassed = [ False for x in result.mQValues ]
        
    result.mPValues = pvalues
    result.mPi0 = pi0
    result.mLambda = vlambda

    return result

def doFDRPython(pvalues, 
                vlambda=None,
                pi0_method="smoother", 
                fdr_level=None, 
                robust=False,
                smooth_df = 3,
                smooth_log_pi0 = False,
                pi0 = None,
                plot = False ):
    """modeled after code taken from http://genomics.princeton.edu/storeylab/qvalue/linux.html.

    I did not like the error handling so I translated most to python.
    
    Compute FDR after method by Storey et al. (2002).
    """

    if min(pvalues) < 0 or max(pvalues) > 1:
        raise ValueError( "p-values out of range" )

    # set to default of qvalue method
    if vlambda == None: vlambda = numpy.arange(0,0.95,0.05)

    m = len(pvalues)
    pvalues = numpy.array( pvalues, dtype = numpy.float )

    if pi0 == None:
        if type(vlambda) == float:
            vlambda = (vlambda,)

        if len(vlambda) > 1 and len(vlambda) < 4:
            raise ValueError(" if length of vlambda greater than 1, you need at least 4 values." )

        if len(vlambda) > 1 and (min(vlambda) < 0 or max(vlambda) >= 1):
            raise ValueError( "vlambda must be within [0, 1).")

        # estimate pi0
        if len(vlambda)==1: 
            vlambda = vlambda[0]
            if  vlambda < 0 or vlambda >=1 :
                raise ValueError( "vlambda must be within [0, 1).")

            pi0 = numpy.mean( [ x >= vlambda for x in pvalues ] ) / (1.0 - vlambda)
            pi0 = min(pi0, 1.0)
        else:

            pi0 = numpy.zeros( len(vlambda), numpy.float )

            for i in range( len(vlambda) ):
                pi0[i] = numpy.mean( [x >= vlambda[i] for x in pvalues ]) / (1.0 - vlambda[i] )

            if pi0_method=="smoother":

                if smooth_log_pi0: pi0 = math.log(pi0)

                tck = scipy.interpolate.splrep( vlambda,
                                                pi0, 
                                                k = smooth_df,
                                                s = 10000 )
                
                if plot:
                    import matplotlib.pyplot as plt
                    plt.figure()
                    plt.plot( vlambda, pi0 )
                    x2 = numpy.arange( 0, 1, 0.001 )
                    y2 = scipy.interpolate.splev( x2, tck )
                    plt.plot( x2, y2 )
                    plt.show()
                
                pi0 = scipy.interpolate.splev( max(vlambda), tck )
                if smooth_log_pi0: pi0 = math.exp(pi0)

            elif pi0_method=="bootstrap":

                minpi0 = min(pi0)

                mse = numpy.zeros( len(vlambda), numpy.float )
                pi0_boot = numpy.zeros( len(vlambda), numpy.float )

                for i in xrange(100):
                    # sample pvalues
                    idx_boot = numpy.random.random_integers( 0, m-1, m) 
                    pvalues_boot = pvalues[idx_boot]

                    for x in xrange( len(vlambda )):
                        # compute number of pvalues larger than lambda[x]
                        pi0_boot[x] = numpy.mean( pvalues_boot > vlambda[x]) / (1.0 - vlambda[x]) 
                    mse += (pi0_boot - minpi0) ** 2
                pi0 = min( pi0[mse==min(mse)] )
            else:
                raise ValueError( "'pi0_method' must be one of 'smoother' or 'bootstrap'.")

            pi0 = min(pi0,1.0)
    
    if pi0 <= 0:
        raise ValueError( "The estimated pi0 <= 0. Check that you have valid p-values or use another vlambda method." )

    if fdr_level != None and (fdr_level <= 0 or fdr_level > 1):
        raise ValueError( "'fdr_level' must be within (0, 1].")

    # compute qvalues
    idx = numpy.argsort( pvalues )
    # monotonically decreasing bins, so that bins[i-1] > x >=  bins[i]
    bins = numpy.unique( pvalues )[::-1]

    # v[i] = number of observations less than or equal to pvalue[i]
    # could this be done more elegantly?
    val2bin = len(bins) - numpy.digitize( pvalues, bins )
    v = numpy.zeros( m, dtype = numpy.int )
    lastbin = None
    for x in xrange( m-1, -1, -1 ):
        bin = val2bin[idx[x]]
        if bin != lastbin: c = x
        v[idx[x]] = c+1
        lastbin = bin

    qvalues = pvalues * pi0 * m / v
    if robust:
        qvalues /= ( 1.0 - ( 1.0 - pvalues)**m )

    # bound qvalues by 1 and make them monotonic
    qvalues[idx[m-1]] = min(qvalues[idx[m-1]],1.0)
    for i in xrange(m-2,-1,-1):
        qvalues[idx[i]] = min(min(qvalues[idx[i]],qvalues[idx[i+1]]),1.0)

    result = FDRResult()
    result.mQValues = qvalues

    if fdr_level != None:
        result.mPassed = [ x <= fdr_level for x in result.mQValues ]
    else:
        result.mPassed = [ False for x in result.mQValues ]
        
    result.mPValues = pvalues
    result.mPi0 = pi0
    result.mLambda = vlambda
    
    result.xvalues = qvalues
    
    return result

#################################################################
#################################################################
#################################################################
class CorrelationTest:
    '''coefficient is r, not r squared'''
    def __init__(self, 
                 r_result = None, 
                 s_result = None,
                 method = None):
        self.mPValue = None
        self.mMethod = None

        if r_result:
            self.mCoefficient = r_result['estimate']['cor']
            self.mPValue = float(r_result['p.value'])
            self.mNObservations = r_result['parameter']['df']
            self.mMethod = r_result['method']
            self.mAlternative = r_result['alternative']
        elif s_result:
            self.mCoefficient = s_result[0]
            self.mPValue = s_result[1]
            self.mNObservations = 0
            self.mAlternative = "two-sided"
        else:
            self.mCoefficient = 0
            self.mPValue = 1
            self.mSignificance = "na"
            self.mNObservations = 0
            self.mAlternative = "na"
            self.mMethod = "na"

        if method: self.mMethod = method

        if self.mPValue != None:
            self.mSignificance = getSignificance( self.mPValue )

    def __str__(self):
        return "\t".join( (
            "%6.4f" % self.mCoefficient,
            "%e" % self.mPValue,
            self.mSignificance,
            "%i" % self.mNObservations,
            self.mMethod,
            self.mAlternative ) )

    @classmethod
    def getHeaders(cls):
        return ("coeff", "pvalue", "significance", "observations", "method", "alternative" )
    
def filterMasked( xvals, yvals, missing = ("na", "Nan", None, ""), dtype = numpy.float ):
    """convert xvals and yvals to numpy array skipping pairs with
    one or more missing values."""
    xmask = [ i in missing for i in xvals ]
    ymask = [ i in missing for i in yvals ]
    return (numpy.array( [xvals[i] for i in range(len(xvals)) if not xmask[i]], dtype = dtype  ),
            numpy.array( [yvals[i] for i in range(len(yvals)) if not ymask[i]], dtype = dtype) )

def doCorrelationTest( xvals, yvals ):
    """compute correlation between x and y.

    Raises a value-error if there are not enough observations.
    """

    if len(xvals) <= 1 or len(yvals) <= 1:
        raise ValueError( "can not compute correlation with no data" )
    if len(xvals) != len(yvals):
        raise ValueError( "data vectors have unequal length" )
    
#     try:
#         result = CorrelationTest( r_result = R.cor_test( xvals, yvals, na_action="na_omit" ) )
#     except rpy.RPyException, msg:
#         raise ValueError( msg )

    x, y = filterMasked( xvals, yvals )

    result = CorrelationTest( s_result = scipy.stats.pearsonr( x, y ),
                              method = "pearson" )
    result.mNObservations = len(x)

    return result

def getPooledVariance( data ):
    """return pooled variance from a
    list of tuples (sample_size, variance)."""
    t, var = 0, 0
    
    for n, s in data:
        t += n
        var += (n-1) * s
    
    assert t > len(data), "sample size smaller than samples combined"

    return var / float(t - len(data))

###################################################################
###################################################################
###################################################################
## compute ROC curves from sorted values
###################################################################
def computeROC( values ):
    '''return a roc curve for *values*. Values
    is a sorted list of (value, bool) pairs.

    Deprecated - use getPerformance instead

    returns a list of (FPR,TPR) tuples.
    '''
    roc = []
    
    npositives = len( [x for x in values if x[1] ] )
    if npositives == 0: 
        raise ValueError( "no positives among values" )

    ntotal = len(values)

    last_value, last_fpr = None, None
    tp, fp = 0, 0
    tn, fn = ntotal - npositives, npositives 

    for value, is_positive in values:
        if is_positive: 
            tp += 1
            fn -= 1
        else: 
            fp += 1
            tn -= 1

        if last_value != value:

            try:
                tpr = float(tp) / (tp + fn)
            except ZeroDivisionError:
                tpr = 0

            try:
                fpr = float(fp) / (fp + tn)
            except ZeroDivisionError:
                fpr = 0
                
            if last_fpr != fpr:
                roc.append( (fpr,tpr) )
                last_fpr = fpr

        last_values = value

    return roc

class TTest:
    def __init__(self): pass

class WelchTTest:
    def __init__(self): pass

PairedTTest = collections.namedtuple( "PairedTTest", "statistic pvalue" )

def doPairedTTest( vals1, vals2) :
    '''perform paired t-test.

    vals1 and vals2 need to contain the same number of elements.
    '''
    return PairedTTest._make( scipy.stats.ttest_rel( vals1, vals2 ) )

def doWelchsTTest(n1, mean1, std1, 
                  n2, mean2, std2,
                  alpha = 0.05 ):
    '''Welch''s approximate t-test for the difference of two means of
    heteroscedasctic populations.

    This functions does a two-tailed test.

    see PMID: 12016052

    :Parameters:
        n1 : int
            number of variates in sample 1
        n2 : int
            number of variates in sample 2
        mean1 : float
            mean of sample 1
        mean2 : float
            mean of sample 2
        std1 : float
            standard deviation of sample 1
        std2 : float
            standard deviation of sample 2
            
    returns a WelchTTest
    '''
    if std1 == 0 and std2 == 0:
        raise ValueError( 'standard deviations are 0.')

    # convert standard deviation to sample variance
    svar1 = std1**2 * n1 / float(n1-1)
    svar2 = std2**2 * n2 / float(n2-1)

    # compute df and test statistic
    df = ((svar1/n1 + svar2/n2)**2) / ( ((svar1/n1)**2)/(n1-1) + ((svar2/n2)**2)/(n2-1))
    denom = numpy.sqrt(svar1/n1+svar2/n2)
    z = abs(mean1 - mean2) / denom

    # do the test
    pvalue = 2 * scipy.stats.t.sf(z,df)
    result = WelchTTest()
    result.mPValue = pvalue 
    result.mDegreesFreedom = df
    result.mZ = z
    result.mMean1 = mean1
    result.mMean2 = mean2
    result.mSampleVariance1 = svar1
    result.mSampleVariance2 = svar2
    result.mDifference = mean1 - mean2
    result.mZLower = scipy.stats.t.ppf( alpha, df )
    result.mZUpper = scipy.stats.t.ppf( 1.0-alpha, df )
    result.mDifferenceLower = result.mZLower * denom
    result.mDifferenceUpper = result.mZUpper * denom

    return result

###################################################################
###################################################################
###################################################################
## 
###################################################################
def getAreaUnderCurve( xvalues, yvalues ):
    '''compute area under curve from a set of discrete x,y coordinates
    using trapezoids.
    
    This is only as accurate as the density of points.
    '''

    assert len(xvalues) == len(yvalues)
    last_x, last_y = xvalues[0], yvalues[0]
    auc = 0
    for x,y in zip(xvalues, yvalues)[1:]:
        dx = x - last_x
        assert not dx <= 0, "x not increasing: %f >= %f" % (last_x, x)
        dy = abs(last_y - y)
        my = min(last_y, y)
        # rectangle plus triangle
        auc += dx * my + dx * dy / 2
        last_x, last_y = x, y
        
    return auc

###################################################################
###################################################################
###################################################################
## 
###################################################################
def getSensitivityRecall( values ):
    '''return sensitivity/selectivity.

    Values is a sorted list of (value, bool) pairs.

    Deprecated - use getPerformance instead
    '''

    npositives = 0.0
    npredicted = 0.0
    l = None
    result = []
    total = float(len(values))
    for value, is_positive in values:
        npredicted += 1.0
        if is_positive > 0: npositives += 1.0
        if value != l:
            result.append( (value, npositives / npredicted, npredicted / total ) )
        l = value
    if l:
        result.append( (l, npositives / npredicted, npredicted/total ) )

    return result

###################################################################
###################################################################
###################################################################
## 
###################################################################
ROCResult = collections.namedtuple( "ROCResult", 
                                    "value pred tp fp tn fn tpr fpr tnr fnr rtpr rfnr" ) 

def getPerformance( values, 
                    skip_redundant = True,
                    false_negatives = False,
                    bin_by_value = True,
                    monotonous = False,
                    multiple = False,
                    increasing = True,
                    total_positives = None,
                    total_false_negatives = None,
                    ):
    '''compute performance estimates for a list of ``(score, flag)``
    tuples in *values*.

    Values is a sorted list of (value, bool) pairs.    

    If the option *false-negative* is set, the input is +/- or 1/0 for a 
    true positive or false negative, respectively.

    TP: true positives
    FP: false positives
    TPR: true positive rate  = true_positives /  predicted
    P: predicted
    FPR: false positive rate = false positives  / predicted
    value: value


    '''

    true_positives = 0
    predicted = 0

    last_value = None
    
    binned_values = []

    for value, flag in values:
        if not bin_by_value:
            if last_value != value:
                binned_values.append( (true_positives, predicted, value) )
        else:
            if last_value != None and last_value != value:
                binned_values.append( (true_positives, predicted, last_value) )

        predicted += 1

        if flag: true_positives += 1

        last_value = value

    binned_values.append( (true_positives, predicted, last_value) )
    binned_values.append( (true_positives, predicted, value) )

    if true_positives == 0:
        raise ValueError("# no true positives!")
        
    if total_positives == None:
        if total_false_negatives:
            positives = float(predicted)
        else:
            positives = float(true_positives)
    else:
        positives = float(total_positives)

    last_positives = None
    last_tpr = None
    result = []

    for true_positives, predicted, value in binned_values:

        if (predicted == 0):
            predicted = 1

        if total_false_negatives:
            false_negatives = predicted - true_positives
            false_positives = 0
            true_negatives = 0
        else:
            true_negatives = 0
            false_negatives = positives - true_positives
            false_positives = predicted - true_positives

        tpr = float(true_positives) / predicted
        fpr = float(false_positives) / (true_positives + false_negatives )
        fnr = float(false_negatives) / positives
        tnr = 0

        # relative rates 
        rfpr = float(false_positives) / predicted
        rfnr = float(false_negatives) / predicted

        if monotonous and last_tpr and last_tpr < tpr:
            continue

        if skip_redundant and true_positives == last_positives:
            continue
        
        if (predicted > 0):
            result.append( ROCResult._make( 
                    (value,
                     predicted,
                     true_positives,
                     false_positives,
                     true_negatives,
                     false_negatives,
                     tpr, fpr, tnr, fnr,
                     rfpr, rfnr ) ) )

        last_positives = true_positives
        last_tpr = tpr

    return result

###################################################################
###################################################################
###################################################################
## 
###################################################################
def doMannWhitneyUTest( xvals, yvals ):
    '''apply the Mann-Whitney U test to test for the difference of medians.'''

    r_result = R.wilcox_test( xvals, yvals, paired = False )

    result = Result().fromR( 
        ( ("pvalue", 'p.value'),
          ('alternative', None),
          ('method', None ) ), 
        r_result )

    return result


###################################################################
###################################################################
###################################################################
## adjust P-Value
###################################################################
def adjustPValues( pvalues, method = 'fdr', n = None ):
    '''returns an array of adjusted pvalues
    
    Reimplementation of p.adjust in the R package.

    p: numeric vector of p-values (possibly with 'NA's).  Any other
    R is coerced by 'as.numeric'.

    method: correction method. Valid values are:

    n: number of comparisons, must be at least 'length(p)'; only set
    this (to non-default) when you know what you are doing
    
    For more information, see the documentation of the
    p.adjust method in R.
    '''

    if n == None: n = len(pvalues)

    if method == "fdr": method = "BH"

    # optional, remove NA values
    p = numpy.array( pvalues, dtype = numpy.float )
    lp = len(p)
    
    assert n <= lp

    if n <= 1:
        return p
    if n == 2 and method == "hommel":
        method = "hochberg"
    if method == "bonferroni":
        p0 = n * p 
    elif method == "holm":
        i = numpy.arange( lp )
        o = numpy.argsort( p )
        ro = numpy.argsort( o )
        m = numpy.maximum.accumulate( (n - i) * p[o] )
        p0 = m[ro]
    elif method == "hommel":
    #     if (n > lp) p <- c(p, rep.int(1, n - lp))
        i = numpy.arange( n )
        o = numpy.argsort( p )
        p = p[o]
        ro = numpy.argsort( o )
        raise NotImplementedError( "hommel method not fully implemented" )
    #     q <- pa <- rep.int(min(n * p/i), n)
    #     for (j in (n - 1):2) {
    #         ij <- seq_len(n - j + 1)
    #         i2 <- (n - j + 2):n
    #         q1 <- min(j * p[i2]/(2:j))
    #         q[ij] <- pmin(j * p[ij], q1)
    #         q[i2] <- q[n - j + 1]
    #         pa <- pmax(pa, q)
    #     }
    #     pmax(pa, p)[if (lp < n) ro[1:lp] else ro]
    elif method == "hochberg":
        i = numpy.arange( 0, lp)[::-1] 
        o = numpy.argsort(1-p)
        ro = numpy.argsort( o )
        m = numpy.minimum.accumulate( (n - i) * p[o] )
        p0 = m[ro]
    elif method == "BH":
        i = numpy.arange( 1, lp + 1)[::-1] 
        o = numpy.argsort(1-p)
        ro = numpy.argsort( o )
        m = numpy.minimum.accumulate( float(n) / i * p[o] )
        p0 = m[ro]
    elif method == "BY":
        i = numpy.arange( 1, lp + 1)[::-1] 
        o = numpy.argsort(1-p)
        ro = numpy.argsort( o )
        q = numpy.sum( 1.0 / numpy.arange( 1, n + 1) )
        m = numpy.minimum.accumulate( q * float(n) / i * p[o])
        p0 = m[ro]
    elif method == "none":
        p0 = p

    return numpy.minimum( p0, numpy.ones( len(p0) ) )




####
####
##
##
## Copyright (C) 2008 Andreas Heger All rights reserved
##
## Author: Andreas Heger <heger@ebi.ac.uk>
##
## $Id: psl2assembly.py 2781 2009-09-10 11:33:14Z andreas $
##
##
####
####
'''
psl2assembly.py - assemble (long) reads on genome
==================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Assemble reads on a reference genome. Overlapping reads are merged.

This script is useful for building transcript models from
long aligned reads (for example, 454 reads or ESTs).

The script is aware of splice motifs and modifies alignments 
around intron boundaries to conform to splice site signals.

Input is a psl formatted file sorted by target contig and target start.

Output is a table linking reads to transcripts. Optional
output files are:

.transcripts: a psl formatted output file containing transcripts

.pileup: fasta file with pileup alignments of reads on transcripts

.regions: a tab separated file of transcript coordinates

.coverage: coverage statistics for each transcript

.sbjct_coverage: coverage statistics of the target genome

Usage
-----

Example::

   python <script_name>.py --help

Type::

   python <script_name>.py --help

for command line help.

Documentation
-------------

Code
----
'''

import sys
import re
import string
import optparse
import time
import os
import tempfile
import shutil
import collections
import math

import scipy
import scipy.stats
import numpy

import CGAT.Intervals as Intervals
import CGAT.Experiment as E
import CGAT.Histogram as Histogram
import CGAT.Blat as Blat
import CGAT.IndexedFasta as IndexedFasta
import alignlib
import CGAT.Mali as Mali
import CGAT.Stats as Stats
import CGAT.Genomics as Genomics



class Processor:

    def __init__(self, genome_fasta, genome_queries, options ):
        self.mGenomeFasta = genome_fasta

        self.mQueriesFasta = queries_fasta

        self.options = options

        # map of id to coverage information
        self.mMapId2Coverage = None

    def setMapId2Coverage( self, d ):
        """set mapping from ids to coverage information from previous run."""
        self.mMapId2Coverage = d

    def getReadCounts( self, matches, lgenome ):
        """return three arrays with the number of reads per position

        The three arrays are:
        1. counts for exons
        2. counts for introns
        3. counts for termini (intron/exon boundaries as well)
        ."""

        is_exon = numpy.zeros( lgenome, numpy.int )
        is_intron = numpy.zeros( lgenome, numpy.int )
        is_terminal = numpy.zeros( lgenome, numpy.int )
        
        ###############################################
        # map matches onto genomic sequence counting
        # introns and exons
        for match in matches:

            if self.options.loglevel >= 4:
                self.options.stdlog.write( "# "+ str( match ) + "\n" )
                self.options.stdlog.flush()

            was_exon = False


            assert lgenome >= match.mMapTarget2Query.getRowTo(), "alignment for match %s is out of bounds (%i > %i)" % \
                (str(match), match.mMapTarget2Query.getRowTo(), lgenome ) 

            # collapse the alignment
            for x in range(match.mMapTarget2Query.getRowFrom(), match.mMapTarget2Query.getRowTo() ):
                if match.mMapTarget2Query.mapRowToCol( x ) >= 0:            
                    is_exon[ x ] += 1
                    if not was_exon: is_terminal[x] += 1
                    was_exon = True
                else:
                    is_intron[ x ] += 1
                    if was_exon: is_terminal[x-1] += 1
                    was_exon = False
                    
            is_terminal[match.mMapTarget2Query.getRowTo()-1] += 1

        return is_exon, is_intron, is_terminal

class Builder(Processor):

    def __init__(self, *args, **kwargs ):
        Processor.__init__( self, *args, **kwargs )

        self.mIdFormat = options.output_format

        if self.options.output_filename_pattern:
            self.mOutFile = open( self.options.output_filename_pattern % self.mName, "w" )
        else:
            self.mOutFile = self.options.stdout

        self.mId = 0

    def __call__( self, id, *args ):

        self.mId = id

        if self.options.loglevel >= 3:
            self.options.stdlog.write("# module %s for %s started\n" % (self.mName, self.mId ) )
            self.options.stdlog.flush()
                                 
        self.mOutputId = self.mIdFormat % self.mId 
        self.process( *args )
        self.mOutFile.flush()

        if self.options.loglevel >= 3:
            self.options.stdlog.write("# module %s for %s finished\n" % (self.mName, self.mId ) )
            self.options.stdlog.flush()

    def printHeader( self ):
        header = self.getHeader()
        if header:
            self.mOutFile.write( header + "\n" )

    def getHeader( self ):
        return None

    def buildPileUp( self, sbjct_id, sbjct_start, sbjct_end, matches, map_genome2transcript = None ):
        """build a pileup multiple alignment of all matches on the genomic sequence.
        
        If map_genome2transcript is given, then the alignment is mapped from genomic coordinates
        (starting a 0) to a different coordinate system. Use this mapper to remove gaps or
        undesired regions from the sequence.
        """

        if self.options.loglevel >= 4:
            self.options.stdlog.write("# pileup started\n" )
            self.options.stdlog.flush()

        mali = alignlib.makeMultAlignment()

        genome = self.mGenomeFasta.getSequence( sbjct_id, "+", sbjct_start, sbjct_end )        

        if map_genome2transcript:
            mapped = alignlib.makeAlignmentBlocks()
            genome = "".join( [ genome[x] for x in range(len(genome)) if map_genome2transcript.mapRowToCol(x) != -1 ] )

        identifiers = ["%s:%s" % (self.mOutputId, "genome" ) ]
        ali = alignlib.makeAlignmentBlocks()
        lgenome = sbjct_end - sbjct_start
        ali.addDiagonal( 0, lgenome, 0 )
        seqs = alignlib.StringVector()
        seqs.append( genome )
        mali.add( ali )

        x = 0

        for match in matches:
            if self.options.loglevel >= 3:
                self.options.stdlog.write( "# "+ str( match ) + "\n" )
            
            query = self.mQueriesFasta.getSequence( match.mQueryId, match.strand )
            
            if self.options.loglevel >= 4:

                a = str(alignlib.AlignmentFormatExplicit( match.mMapTarget2Query,
                                                          alignlib.makeSequence( genome ),
                                                          alignlib.makeSequence( query ) ))

                # self.options.stdlog.write( str( alignlib.AlignmentFormatExplicit( match.mMapTarget2Query, 
#                                                                                  alignlib.makeSequence( genome ),
#                                                                                  alignlib.makeSequence( query ) ) ) + "\n" )
            
            if map_genome2transcript:
                alignlib.combineAlignment( mapped, 
                                           map_genome2transcript,
                                           match.mMapTarget2Query,
                                           alignlib.RR )
                                            
                mali.add( mapped )
            else:
                mali.add( match.mMapTarget2Query )
            
            seqs.append( query )

            x += 1
            identifiers.append( "%s:%s" % (self.mOutputId, match.mQueryId ))
        
        if self.options.loglevel >= 4:
            self.options.stdlog.write("# pileup finished\n" )
            self.options.stdlog.flush()

        return mali, identifiers, seqs 
    
    def finish( self ):
        self.options.stdlog.write( "# %s\tfinished\n" % (self.mName ))

class BuilderRegion(Builder):

    mName = "regions"

    def __init__(self, *args, **kwargs ):
        Builder.__init__( self, *args, **kwargs )

    def getHeader( self ):
        return "id\tcontig\tstart\tend\tsize\tnmatches\tnpos\tnneg"

    @E.benchmark
    def process( self, sbjct_id, sbjct_start, sbjct_end, matches ):
        npositive = sum( Genomics.IsPositiveStrand( x.strand ) for x in matches )
        nnegative = sum( Genomics.IsNegativeStrand( x.strand ) for x in matches )
        self.mOutFile.write( "%s\t%s\t%i\t%i\t%i\t%i\t%i\t%i\n" % (self.mOutputId, sbjct_id, sbjct_start, sbjct_end, sbjct_end - sbjct_start, len(matches), npositive, nnegative ))
        self.mOutFile.flush()


class BuilderTranscribedLocus(Builder):

    mName = "locus.psl"

    """build a transcript from all aligned matches."""
    def __init__(self, *args, **kwargs ):
        Builder.__init__( self, *args, **kwargs )

        if self.options.output_filename_pattern:
            self.mOutFileFasta = open( self.options.output_filename_pattern % self.mName + ".fasta", "w" )
            if self.options.output_pileup:
                self.mOutFilePileup = open( self.options.output_filename_pattern % self.mName + ".pileup", "w" )
            self.mOutFileStrand = open( self.options.output_filename_pattern % self.mName + ".strand", "w" )
        else:
            self.mOutFileFasta = self.options.stdout
            self.mOutFileStrand = self.options.stdout
            if self.options.output_pileup:
                self.mOutFilePileup = self.options.stdout

        self.mOutFileStrand.write( "gene_id\tstrand\tnmatches\tnpos\tnneg\tnintrons\tnintrons_pos\tnintrons_neg\tnintrons_unknown\n" )

    def getHeader( self ):
        return Blat.Match().getHeader()

    def guessStrand( self, matches, intron_counts = None ):
        """guess the strand of a transcript.

        The first guess is based on the directionality of splice site motifs
        given by the dict intron_counts. If that is inconclusive, the directionality
        of the matches is considered.

        Only clear votes count, i.e. if one splice site/read disagrees, the 
        strand is labelled ambiguous "."
        """
        strand = None
        if intron_counts and sum( intron_counts.values() ) > 0:
            nintrons_positive = intron_counts["+"]
            nintrons_negative = intron_counts["-"]
            nintrons_unknown = intron_counts["."]

            if nintrons_positive and nintrons_negative == 0:
                strand = "+"
            elif nintrons_negative and nintrons_positive == 0:
                strand = "-"
            else:
                strand = "."
        else:
            nintrons_positive = 0
            nintrons_negative = 0
            nintrons_unknown = 0

        npositive = sum( Genomics.IsPositiveStrand( x.strand ) for x in matches )
        nnegative = sum( Genomics.IsNegativeStrand( x.strand ) for x in matches )

        if not strand:
            if npositive and nnegative == 0:
                strand = "+"
            elif nnegative and npositive == 0:
                strand = "-"
            else:
                strand = "."

        self.mOutFileStrand.write( "%s\t%s\t%i\t%i\t%i\t%i\t%i\t%i\t%i\n" % \
                                        (self.mOutputId, 
                                         strand,
                                         len(matches),
                                         npositive,
                                         nnegative,
                                         nintrons_positive + nintrons_negative + nintrons_unknown,
                                         nintrons_positive,
                                         nintrons_negative,
                                         nintrons_unknown ) )

        return strand

    def getTranscriptMap( self, sbjct_id, sbjct_start, sbjct_end, matches ):
        
        genome = self.mGenomeFasta.getSequence( sbjct_id, "+", sbjct_start, sbjct_end )
        
        residues = set()
        for match in matches:
            if self.options.loglevel >= 3:
                self.options.stdlog.write( "# "+ str( match ) + "\n" )
                self.options.stdlog.flush()

            # collapse the alignment
            for x in range(match.mMapTarget2Query.getRowFrom(), match.mMapTarget2Query.getRowTo() ):
                if match.mMapTarget2Query.mapRowToCol( x ) >= 0:            
                    residues.add( x )
        
        residues = list( residues )
        residues.sort()
        
        map_transcript2genome = alignlib.makeAlignmentBlocks()

        sequence = []

        for x in range(len(residues)):
            sequence.append( genome[residues[x]] )
            map_transcript2genome.addPair( x, residues[x] )
            
        map_transcript2genome.moveAlignment( 0, sbjct_start )

        return map_transcript2genome, "".join(sequence ), self.guessStrand( matches )

    @E.benchmark
    def process( self, sbjct_id, sbjct_start, sbjct_end, matches ):

        map_transcript2genome, sequence, strand = self.getTranscriptMap( sbjct_id, sbjct_start, sbjct_end, matches )

        match = Blat.Match()
        match.fromMap( map_transcript2genome )
        ## ambiguous strands are labelled "+" as psl does not know
        if strand == "-":
            match.strand = "-"
        else:
            match.strand = "+"

        match.mSbjctId = sbjct_id
        match.mQueryId = self.mOutputId
        match.mQueryLength = map_transcript2genome.getRowTo()
        match.mSbjctLength = self.mGenomeFasta.getLength( sbjct_id )

        self.mOutFile.write( str( match ) + "\n" )
        self.mOutFile.flush()
        self.mOutFileFasta.write( ">%s\n%s\n" % (self.mOutputId, sequence) )
        self.mOutFileFasta.flush()

        if self.options.output_pileup:
            self.outputPileUp( sbjct_id, sbjct_start, sbjct_end, matches, match, sequence )

    def outputPileUp( self, sbjct_id, sbjct_start, sbjct_end, matches, match, sequence ):
        """output a pileup multiple alignment for debugging purposes.

        Matches: the reads
        match: the transcript match
        """

        mali, identifiers, seqs = self.buildPileUp( sbjct_id, sbjct_start, sbjct_end, matches )
        
        map_genome2query = match.getMapTarget2Query()
        map_genome2query.moveAlignment( -sbjct_start, 0 ) 
        
        mali.add( map_genome2query )
        seqs.append( sequence )

        identifiers.append( "%s:%s" % (self.mOutputId, "transcript" ))

        if options.mali_output_format == "fasta":
            mm = Mali.convertAlignlib2Mali( mali, identifiers, seqs )
            mm.writeToFile( self.mOutFilePileup, format = "fasta" )

class BuilderTranscriptSpliced(BuilderTranscribedLocus):

    mName = "transcripts.psl"

    """build a transcript from all aligned matches.

    This processor includes a heuristics to correct for errors
    in the alignment/sequencing stage. The following errors are
    corrected:

    1. small gaps in the transcript. These are filled with genomic sequence
    2. splice site overrun = transcripts ending close to a splice site sometimes
       overshoot the splice site.

       Check for the existence of a splice site close by that is supported by at least one
       read.
       
    """
    mMaxGapLength = 15
    mFixIntronDistance = 10
    ## intron motifs [ forward strand ]
    mIntronTypes = ( ( "U2-GT/AG", "GT", "AG"),
                     ( "U2-nc-GC/AG", "GC", "AG"),
                     ( "U12-AT/AC", "AT", "AC") )
    mMaxSpliceMotifLength = 2

    def __init__(self, *args, **kwargs ):
        BuilderTranscribedLocus.__init__( self, *args, **kwargs )

        self.mFixedOverruns = 0
        self.mFixedGaps = 0
        self.mFixedCoding = 0
        self.mNIntrons = 0

        self.mIntronCounts = collections.defaultdict( int )

    def getIntronType( self, s ):
        """return intron type for an intronic sequence. 

        Checks both directions and returns the type of the intron and the strand.
        If the intron-type is not known, type is set to "unknown"
        """

        r = string.translate( s, string.maketrans("ACGTacgt", "TGCAtgca") )[::-1]

        for name, prime5, prime3 in self.mIntronTypes:
            if s[:len(prime5)].upper() == prime5 and \
                    s[-len(prime3):].upper() == prime3:
                return name, "+"
            elif r[:len(prime5)].upper() == prime5 and \
                    r[-len(prime3):].upper() == prime3:
                return name, "-"
        else:
            return "unknown", "."

    def getTranscriptMap( self, sbjct_id, sbjct_start, sbjct_end, matches ):

        genome = self.mGenomeFasta.getSequence( sbjct_id, "+", sbjct_start, sbjct_end ).upper()

        lgenome = sbjct_end - sbjct_start
        is_exon, is_intron, is_terminal = self.getReadCounts( matches, lgenome )

        ###############################################
        # fill small gaps
        is_gap = False
        lgap = 0
        for x in xrange( 0, lgenome ):
            if is_exon[x] == 0:
                lgap += 1
            else:
                if lgap:
                    if lgap <= self.mMaxGapLength:
                        if self.options.loglevel >= 3:
                            self.options.stdlog.write( "# fixing small gap of length %i: %i-%i\n" % ( lgap, x-lgap, x) )

                        for y in xrange( x - lgap, x):
                            is_exon[y] = 1
                            is_intron[y] = 0
                        self.mFixedGaps += 1

                lgap = 0

        
        ###############################################
        # deal with intron overrun
        def hasSpliceMotif( sequence ):
            """find a splice motif in seq."""
            s = sequence.upper()
            r = string.translate( s.upper(), string.maketrans("ACGTacgt", "TGCAtgca") )[::-1]

            for name, prime5, prime3 in self.mIntronTypes:
                if s.startswith( prime5 ) or r.startswith( prime5 ) or \
                        s.startswith( prime3 ) or r.startswith( prime3 ):
                    return True
            return False

        def findSpliceStart( start, end ):
            """find a splice site start within start,end with more evidence than the original site."""
            for x in range(end,start,-1):
                if is_intron[x] > is_intron[x-1] and hasSpliceMotif( genome[x:x+self.mMaxSpliceMotifLength] ):
                    return x
            return None

        def findSpliceEnd( start, end ):
            """find a splice site end within start,end with more evidence than the original site."""
            for x in range(start,end):
                if is_intron[x-1] > is_intron[x] and hasSpliceMotif( genome[x-self.mMaxSpliceMotifLength:x] ):
                    return x-1
            return None

        for x in xrange( 0, lgenome ):
            # check for read ends within introns
            if self.options.loglevel >= 5:
                self.options.stdlog.write( "# %i(%i): is_term=%i, is_exon=%i, is_intron=%i, check=%i\n" %\
                                               (x, x+sbjct_start,
                                                is_terminal[x], is_exon[x], is_intron[x],
                                                is_terminal[x] and is_terminal[x] == is_exon[x] and is_intron[x] ) )

            if is_terminal[x] and is_terminal[x] == is_exon[x] and is_intron[x]:
                left = max(0, x-self.mFixIntronDistance )
                right = min( lgenome-1, x+self.mFixIntronDistance )
                if is_exon[left] > 0:
                    if is_exon[left] > is_exon[x]:
                        z = findSpliceStart( left, x )
                        if z != None:
                            if self.options.loglevel >= 3:
                                self.options.stdlog.write( "# correcting intron boundary: %i-%i %s %s %s\n" %\
                                                               ( z, 
                                                                 x + 1, 
                                                                 genome[z-5:z],
                                                                 genome[z:x+1],
                                                                 genome[x+1:x+6]))

                            for y in range( z, x + 1):
                                is_exon[y] = 0
                                is_intron[y] = 1
                            self.mFixedOverruns += 1

                elif is_exon[right] > 0:
                    if is_exon[right] > is_exon[x]:
                        z = findSpliceEnd( x, right )
                        if z != None:
                            if self.options.loglevel >= 3:
                                self.options.stdlog.write( "# correcting intron boundary: %i-%i: %s %s %s\n" %\
                                                           ( x, z + 1, 
                                                             genome[x-5:x],
                                                             genome[x:z+1],
                                                             genome[z+1:z+6]))

                            for y in range( x, z + 1):
                                is_exon[y] = 0
                                is_intron[y] = 1
                            self.mFixedOverruns += 1
                
        map_transcript2genome = alignlib.makeAlignmentBlocks()

        sequence = []

        y = 0
        for x in range(len(is_exon)):
            if is_exon[x]:
                sequence.append( genome[x] )
                map_transcript2genome.addPair( y, x )
                y += 1

        map_transcript2genome.moveAlignment( 0, sbjct_start )

        ## decide upon strand based on introns
        ## and do counting
        strands = { ".": 0, "+" : 0, "-": 0 }
        for start, end in self.getIntrons( is_exon ):
            name, strand = self.getIntronType( genome[start:end] )
            strands[strand] += 1
            if options.loglevel >= 3:
                self.options.stdlog.write( "# intron check: %i-%i: %i-%i: %s %s\n" %\
                                               ( start, end, 
                                                 start+sbjct_start,
                                                 end+sbjct_start,
                                                 name, strand ) )


            self.mIntronCounts[name] += 1
            
        return map_transcript2genome, "".join(sequence), self.guessStrand( matches, intron_counts = strands )

    def getIntrons( self, is_exon ):
        """return a list of intron intervals."""
        introns = []
        l = len(is_exon)
        start, end = 0, l - 1
        while start < l and is_exon[start] == 0: start += 1
        while end >= 0 and is_exon[end] == 0: end -= 1
            
        was_intron = False
        f = 0
        for x in xrange( start, end ):
            if is_exon[x] and was_intron:
                introns.append( (f, x) )
            elif not is_exon[x] and not was_intron:
                f = x
            was_intron = not is_exon[x]

        if was_intron: introns.append( (f,end) )
        return introns

    def finish( self ):

        if self.options.loglevel >= 1:
            self.options.stdlog.write( "# %s\tfinished: fixed_gaps=%i, fixed_overruns=%i\n" % (self.mName, self.mFixedGaps, self.mFixedOverruns ) )
            total = sum( self.mIntronCounts.values() )
            if total:
                self.options.stdlog.write( "# %s\tstats:\n" % self.mName )
                self.options.stdlog.write( "# %s\ttype\tcounts\tfreq\n" % (self.mName) )
                for key,val in sorted(self.mIntronCounts.items()):
                    self.options.stdlog.write( "# %s\t%s\t%i\t%5.2f\n" % (self.mName, key, val, 100.0 * val / total ) )
                self.options.stdlog.write( "# %s\t%s\t%i\t%5.2f\n" % ( self.mName, "total", total, 100.0 * total / total ) )

class BuilderCoverage(Builder):
    """compute residue read coverage over all genomic bases that have at least one aligned
    match.

    If previous coverage information is available, aggregate information is output.
    (updated mean and pooled standard deviation).
    """

    mName = "coverage"

    def __init__(self, *args, **kwargs ):
        Builder.__init__( self, *args, **kwargs )
       
    def getHeader( self ):
        ## if using previous coverage values, only mean is available
        if self.mMapId2Coverage:
            return "id\tcontig\tstart\tend\tsize\tnmatches\tncovered\tmean\tstddev" 
        else:
            return "id\tcontig\tstart\tend\tsize\tnmatches\tncovered\t%s" % ("\t".join(Stats.DistributionalParameters().getHeaders()) )

    @E.benchmark
    def process( self, sbjct_id, sbjct_start, sbjct_end, matches ):

        residues = collections.defaultdict(int)
        for match in matches:
            if self.options.loglevel >= 4:
                self.options.stdlog.write( "# "+ str( match ) + "\n" )
            
            # collapse the alignment
            # offset = match.mSbjctFrom - sbjct_start
            for x in range(match.mMapTarget2Query.getRowFrom(), match.mMapTarget2Query.getRowTo() ):
                if match.mMapTarget2Query.mapRowToCol( x ) >= 0:            
                    residues[x] += 1

        ncovered = len(residues)

        if self.mMapId2Coverage:
            c, t, nmatches = 0, 0, 0

            vars = []
            for match in matches:
                try:
                    m = self.mMapId2Coverage[match.mQueryId]
                except KeyError:
                    E.warn("no coverage information for %s" % match.mQueryId)
                    continue
                c += m.mMean * m.mNCovered
                t += m.mNCovered
                nmatches += m.mNMatches
                vars.append( (m.mNCovered, m.mStddev * m.mStddev)  )
                
            pooled_variance = Stats.getPooledVariance( vars )
            extra_info = "%6.4f\t%6.4f" % ( (float(c) / t), math.sqrt( pooled_variance) ) 

        else:
            nmatches = len(matches)
            extra_info = str(Stats.DistributionalParameters( residues.values() ))

        self.mOutFile.write( "%s\t%s\t%i\t%i\t%i\t%i\t%i\t%s\n" % (self.mOutputId, sbjct_id, sbjct_start, sbjct_end, 
                                                                   sbjct_end - sbjct_start, 
                                                                   nmatches,
                                                                   ncovered,
                                                                   extra_info))
        self.mOutFile.flush()



class BuilderSbjctCoverage(Builder):
    """compute residue read coverage over all genomic bases in the sbjct
    irrespective if they match zero or more reads.
    """

    mName = "sbjct_coverage"

    def __init__(self, *args, **kwargs ):
        Builder.__init__( self, *args, **kwargs )
       
    def getHeader( self ):
        return "id\tcontig\tstart\tend\tsize\tnmatches\tncovered\t%s" % ("\t".join(Stats.DistributionalParameters().getHeaders()) )

    @E.benchmark
    def process( self, sbjct_id, sbjct_start, sbjct_end, matches ):

        sbjct_length = self.mGenomeFasta.getLength( sbjct_id )

        residues = collections.defaultdict(int)

        for match in matches:
            if self.options.loglevel >= 4:
                self.options.stdlog.write( "# "+ str( match ) + "\n" )
            
            # collapse the alignment
            # offset = match.mSbjctFrom - sbjct_start
            for x in range(match.mMapTarget2Query.getRowFrom(), match.mMapTarget2Query.getRowTo() ):
                if match.mMapTarget2Query.mapRowToCol( x ) >= 0:            
                    residues[x] += 1

        self.mOutFile.write( "%s\t%s\t%i\t%i\t%i\t%i\t%i\t%s\n" % (self.mOutputId, sbjct_id, sbjct_start, sbjct_end, 
                                                                   sbjct_end - sbjct_start, 
                                                                   len(matches),
                                                                   len(residues),
                                                                   str(Stats.DistributionalParameters( residues.values() + [0.0] * (sbjct_length - len(residues)) ) )))
        self.mOutFile.flush()

class BuilderIndels(Builder):
    """indel variation in the genome or the reads.
    """

    mName = "indels"

    # region to examine on either side to call an indel in the genome
    mRegionSize = 5

    # minimum support (= number of reads) on either side of an indel
    mMinSupport = 3

    def __init__(self, *args, **kwargs ):
        Builder.__init__( self, *args, **kwargs )

        assert self.mQueriesFasta, "%s requires a sequence database with the queries" % self.mName
        assert self.mGenomeFasta, "%s requires a sequence database with the genome" % self.mName
       
        self.mNDeletions = 0
        self.mNInsertions = 0

        if self.options.loglevel >= 1:
            self.options.stdlog.write("# %s\tstarted: mRegionSize=%i, mMinSupport=%i\n" % (self.mName, self.mRegionSize, self.mMinSupport) )

    def getHeader( self ):
        return "id\tcontig\tpos\tposT\tposM\tchar\tnmatches\tnaligned\tngaps\tsupport"

    @E.benchmark
    def process( self, sbjct_id, sbjct_start, sbjct_end, matches ):
        """output info on positions that are gap-chars in genome.

        This script will only report single base indels.
        """

        lgenome = sbjct_end - sbjct_start
        is_exon, is_intron, is_terminal = self.getReadCounts( matches, lgenome )
        mali, identifiers, seqs = self.buildPileUp( sbjct_id, sbjct_start, sbjct_end, matches )

        if self.options.loglevel >= 4:
            self.options.stdlog.write("# converting mali started\n" )
            self.options.stdlog.flush()
        
        columns = Mali.convertAlignlib2Mali( mali, identifiers, seqs ).getColumns()

        if self.options.loglevel >= 4:
            self.options.stdlog.write("# converting finished\n" )
            self.options.stdlog.flush()

        assert identifiers[0] == "%s:genome" % self.mOutputId

        nmatches = len(matches)

        min_support = self.mMinSupport

        # c: index in alignment, g: index on genome
        # skip first and last position
        c, g = 1, 1
        for column in columns[1:-1]:

            ngaps = column[1:].count("-")
            # can not use is_exons, as this has only positions on genome
            naligned = nmatches - ngaps
            is_gap = column[0] == "-"
            # deletion in genome: support = reads with inserted base
            if (is_gap and naligned >= min_support) or \
                    (not is_gap and is_exon[g-1] >= min_support and naligned == 0 and is_exon[g+1] >= min_support):

                naligned_left = min( is_exon[ max(0, g-self.mRegionSize):g] )
                naligned_right = min( is_exon[ g+1:min(lgenome, g+self.mRegionSize+1)] ) 
                support = min( (naligned_left, naligned_right) )

                if support >= min_support:
                    if is_gap:
                        self.mNDeletions += 1
                    else:
                        self.mNInsertions += 1

                    self.mOutFile.write( "%s\t%s\t%i\t%i\t%i\t%s\t%i\t%i\t%i\t%i\n" % ( self.mOutputId, 
                                                                                    sbjct_id, 
                                                                                    sbjct_start + g, 
                                                                                    g,
                                                                                    c,
                                                                                    column[0],
                                                                                    nmatches,
                                                                                    naligned,
                                                                                    ngaps,
                                                                                    support ) )
                    self.mOutFile.flush()

            # increment indices: skip over gaps in genome
            if not is_gap: g += 1
            c += 1

    def finish( self ):

        if self.options.loglevel >= 1:
            self.options.stdlog.write( "# %s\tfinished: ndeletions=%i, ninsertions=%i\n" % (self.mName, self.mNDeletions, self.mNInsertions ) )

##-------------------------------------------------------------------------
class BuilderConsensus(Builder):
    """builds a pileup alignment of reads on the transcript sequence 
    and outputs the consensus sequence (without introns).

    This module can be used to build transcripts of cross-species
    mappings. Note that the sequence will no correspond to the
    one output by BuilderTranscriptsIntrons as it does not 
    correct for exon overrun.
    """
    
    mName = "consensus"

    def __init__(self, *args, **kwargs ):
        Builder.__init__( self, *args, **kwargs )

        assert self.mQueriesFasta, "%s requires a sequence database with the queries" % self.mName
        assert self.mGenomeFasta, "%s requires a sequence database with the genome" % self.mName

    @E.benchmark
    def process( self, sbjct_id, sbjct_start, sbjct_end, matches ):

        lgenome = sbjct_end - sbjct_start
        is_exon, is_intron, is_terminal = self.getReadCounts( matches, lgenome )
        
        map_genome2transcript = alignlib.makeAlignmentVector()

        c = 0
        for x in range(lgenome):
            if is_exon[x] > 0:
                map_genome2transcript.addPair( x, c, 0 )
                c+=1
                
        if options.loglevel >= 3:
            options.stdlog.write( "# %s\n" % str(alignlib.AlignmentFormatEmissions( map_genome2transcript)) )
            options.stdlog.flush()

        mali, identifiers, seqs = self.buildPileUp( sbjct_id, sbjct_start, sbjct_end, matches,
                                                    map_genome2transcript = map_genome2transcript)

        if options.loglevel >= 4:
            print str(mali)
            options.stdlog.flush()

        if not identifiers:
            identifiers = [ "%i" % x for x in range( mali.getNumSequences() ) ]

        if options.mali_output_format == "fasta":
            mm = Mali.convertAlignlib2Mali( mali, identifiers, seqs )
            mm.writeToFile( self.mOutFile, format = "fasta" )

##-------------------------------------------------------------------------
class BuilderPileUp(Builder):

    mName = "pileup"

    def __init__(self, *args, **kwargs ):
        Builder.__init__( self, *args, **kwargs )

        assert self.mQueriesFasta, "%s requires a sequence database with the queries" % self.mName
        assert self.mGenomeFasta, "%s requires a sequence database with the genome" % self.mName

    @E.benchmark
    def process( self, sbjct_id, sbjct_start, sbjct_end, matches ):

        mali, identifiers, seqs = self.buildPileUp( sbjct_id, sbjct_start, sbjct_end, matches )

        if not identifiers:
            identifiers = [ "%i" % x for x in range( mali.getNumSequences() ) ]

        if options.mali_output_format == "fasta":
            mm = Mali.convertAlignlib2Mali( mali, identifiers, seqs )
            mm.writeToFile( self.mOutFile, format = "fasta" )


##-------------------------------------------------------------------------
class BuilderPolyA(Builder):
    """predict polyA tails.

    The output is a tab-separated table with the following columns

    self.mOutFile.write( "id\ttails\tnremoved\tnstrands\tnmotifs\tninconsistent\n" )

    id
       transcript id
    tails
       number of predicted tails    
    nremoved
       number of tails that were joined with other tails
    nstrands
       number of strands that tails are pointing to
    nmotifs
       number of detected polyA motifs (AATAAA)
    ninconsistent
       number of inconsistent tails

    """    
    mName = "polyA"

    mMotifs = { "AATAAA": ("AATAAA", "TTTATT") }

    # number of residues to search for polyA motifs
    mMotifArea = 50

    # combine polyA starts that are closer than x
    mMaxAggregateArea = 10

    # max residues missing from non polyA end
    mMaxUnaligned = 3
    # min residues in tail
    mMinUnaligned = 10
    # min percent residues that are A/T in tail
    mMinPercent = 70.0

    def __init__(self, *args, **kwargs ):
        Builder.__init__( self, *args, **kwargs )

        assert self.mQueriesFasta, "%s requires a sequence database with the queries" % self.mName
        assert self.mGenomeFasta, "%s requires a sequence database with the genome" % self.mName

        self.mOutFile.write( "id\ttails\tnremoved\tnstrands\tnmotifs\tninconsistent\n" )

        self.mOutFileTails = open( self.options.output_filename_pattern % self.mName + ".tails", "w" )
        self.mOutFileTails.write( "id\tpos\tstrand\tnmotifs\tmotifs\tnids\tids\n" )

    @E.benchmark        
    def process( self, sbjct_id, sbjct_start, sbjct_end, matches ):
        """detect PolyA tails in transcripts.

        1. collect all matches with polyA tail - the remainder is unchanged
        2. declare polyA tail by the shortest! unaligned segment and compute
        coverage for each match appropriately. 
        
        The method checks whether the tails are consistent (always at the same end).
        If not, an AssertionError is thrown
        """

        genome = self.mGenomeFasta.getSequence( sbjct_id, "+", sbjct_start, sbjct_end ).upper()

        tails = collections.defaultdict( list )

        for match in matches:

            if options.loglevel >= 5:
                options.stdlog.write( "#%s\n" % str( match) )

            query_id = match.mQueryId
            missing_start = match.mQueryFrom
            missing_end = match.mQueryLength - match.mQueryTo
            if missing_start < missing_end:
                smaller = missing_start
                larger = missing_end
                start,end = match.mQueryTo,match.mQueryLength
                if match.strand == "+":
                    poly_start = match.mQueryTo - 1
                    direction = "+"
                else:
                    poly_start = match.mQueryLength - match.mQueryTo
                    direction = "-"
            else:
                smaller = missing_end
                larger = missing_start
                start,end = 0, match.mQueryFrom
                if match.strand == "+":
                    poly_start = match.mQueryFrom
                    direction = "-"
                else:
                    poly_start = match.mQueryLength - match.mQueryFrom - 1
                    direction = "+"

            # check if tail is at least polyA_min_aligned and at most polyA_max_unaligned
            # are missing from the other end.
            if not(smaller < self.mMaxUnaligned and larger > self.mMinUnaligned):
                continue

            tail = queries_fasta.getSequence( query_id )[start:end]

            counts = {"A": 0, "T": 0, "N": 0}
            for c in tail.upper(): counts[c] = counts.get(c, 0) + 1
            total = end-start
            pA = 100.0 * (counts["A"] + counts["N"]) / total
            pT = 100.0 * (counts["T"] + counts["N"]) / total
            
            if max(pA,pT) < self.mMinPercent:
                continue

            map_query2genome = match.getMapQuery2Target()
            map_query2genome.moveAlignment( 0, -sbjct_start ) 
            
            mapped = map_query2genome.mapRowToCol( poly_start )

            tails[(mapped,direction)].append( (query_id,len(tail)) )

            if options.loglevel >= 5:
                options.stdlog.write( "# polyA detection: found=%s:%i-%i pA=%5.2f pT=%5.2f pos=%i:%s tail=%i:%s\n" % (query_id, start, end, pA, pT, mapped, direction,len(tail),tail))
                
        if not tails: 
            self.mOutFile.write( "%s\t%i\t%i\t%i\t%i\t%i\n" % \
                                     (self.mOutputId,
                                      len(tails),
                                      0, 0, 0, 0))
            return

        # filter polyA tails
        if options.loglevel >= 3:
            options.stdlog.write( "# putative polyA tails: %s\n" % (str(tails)))

        # aggregate tails
        k = sorted(tails.keys())
        data = tails[k[0]]
        last_mapped, last_strand = k[0]
        new_tails = { (last_mapped, last_strand): data }
        for this_key in k[1:]:
            mapped,strand =this_key

            if strand == last_strand and mapped - last_mapped < self.mMaxAggregateArea:
                data += tails[ this_key ]
            else:
                data = tails[ this_key]
                new_tails[ this_key ] = data

            last_strand, last_mapped = strand, mapped

        nreduced = len(tails) - len(new_tails)
        tails = new_tails

        # check direction
        directions = set([ x[1] for x in tails.keys() ])

        if len(directions) > 1:

            # use number of residues in tail to decide upon strand
            #nforward = sum([ sum([ y[1] for y in x[1]]) for x in tails.items() if x[0][1] == "+"] )
            #nreverse = sum([ sum([ y[1] for y in x[1]]) for x in tails.items() if x[0][1] == "-"] )
            nforward = sum([ len(x[1]) for x in tails.items() if x[0][1] == "+"] )
            nreverse = sum([ len(x[1]) for x in tails.items() if x[0][1] == "-"] )

            ninconsistent = min(nforward, nreverse )

            if options.loglevel >= 1:
                options.stdlog.write("# WARNING: %s: inconsistent polyA tails: forward=%i, reverse=%i from %s\n" % \
                                         (self.mOutputId,
                                          nforward, nreverse, str(tails)) )

            #if nforward >= nreverse:
            #    tails = dict( [ (x[0],x[1]) for x in tails.items() if x[0][1] == "+"] )
            #else:
            #    tails = dict( [ (x[0],x[1]) for x in tails.items() if x[0][1] == "-"] )
            
        else:
            ninconsistent = 0

        nmotifs = 0
        for key, ids in tails.items():

            mapped, direction = key

            distances = []
            for name, motifs in self.mMotifs.items():
                if direction == "+":
                    pos = genome.rfind( motifs[0], mapped-self.mMotifArea,mapped )
                else:
                    pos = genome.find( motifs[1], mapped, mapped+self.mMotifArea )

                if pos >= 0:
                    distances.append( (name, abs(pos - mapped)) )
            
            self.mOutFileTails.write( "%s\t%i\t%s\t%i\t%s\t%s\t%s\n" % \
                                          (self.mOutputId,
                                           mapped,
                                           direction,
                                           len(distances),
                                           ",".join( [ "%s:%i" % x for x in distances ] ),
                                           len(ids),
                                           ",".join( [ "%s:%i" % x for x in ids] )))

            if len(distances): nmotifs += 1

        self.mOutFile.write( "%s\t%i\t%i\t%i\t%i\t%i\n" % \
                                 (self.mOutputId,
                                  len(tails),
                                  nreduced,
                                  len(directions),
                                  nmotifs,
                                  ninconsistent))
        

        self.mOutFileTails.flush()
        self.mOutFile.flush()

##-------------------------------------------------------------------------
class Filter(Processor):

    mName = "Filter"

    def __init__(self, *args, **kwargs ):
        Processor.__init__( self, *args, **kwargs )

        if self.options.output_filename_pattern:
            self.mOutFile = open( self.options.output_filename_pattern % self.mName, "w" )
        else:
            self.mOutFile = self.options.stdout

    def __call__( self, *args ):
        return self.filter( *args )

class FilterImperfect(Filter):
    """remove imperfect matches, if there is no evidence from other matches."""

    mName = "FilterImperfect"

    mMaxNMismatches = 0
    mMaxNMissingResidues = 0

    mMaxQueryNGapsCounts = 0
    mMaxQueryNGapsBases = 0
    mMaxSbjctNGapsCounts = 0
    mMaxSbjctNGapsBases = 0

    def __init__(self, *args, **kwargs ):
        Filter.__init__( self, *args, **kwargs )

    def filter( self, matches ):
        """filter matches."""
        
        if self.options.loglevel >= 2:
            options.stdlog.write( "# started filter %s with %i matches\n" %\
                                      (self.mName, len(matches) ) )
            options.stdlog.flush()

        if len(matches) == 1:
            match = matches[0]
            if match.mQueryLength - (match.mQueryTo - match.mQueryFrom) > self.mMaxNMissingResidues or \
                    match.mNMismatches > self.mMaxNMismatches or \
                    match.mQueryNGapsCounts > self.mMaxQueryNGapsCounts or \
                    match.mQueryNGapsBases > self.mMaxQueryNGapsBases or \
                    match.mSbjctNGapsCounts > self.mMaxSbjctNGapsCounts or \
                    match.mSbjctNGapsBases > self.mMaxSbjctNGapsBases:
                return []
            
        return matches

class FilterIntrons(Filter):
    """remove intronic matches that have little evidence

    An intron needs to be supported by at least two matches and
    the intron boundaries need to coincide perfectly.

    Introns need to be at least mIntronSize residues long.
    """

    # minimum intron size
    mIntronSize = 30

    def __init__(self, *args, **kwargs ):
        Filter.__init__( self, *args, **kwargs )

    def filter( self, matches ):
        """filter matches.
        not fully implemented.
        """
        
        raise "not implemented"
        matches_without_introns, matches_with_introns = [], []

        for match in matches:
            if match.mSbjctNGapsBases < self.mIntronSize:
                matches_without_introns.append( match )
            else:
                # ignore matches with two introns or two introns and a gap
                if match.mSbjctNGapsCounts <= 1:
                    matches_with_introns.append( match )
            
        new_matches_with_introns = []
        for match in matches_with_introns:

            ## find intron
            b = match.mBlockSizes[0] 
            last_q = match.mQueryBlockStarts[0] + b
            last_s = match.mSbjctBlockStarts[0] 
            for q, s, b in zip( match.mQueryBlockStarts[1:], match.mSbjctBlockStarts[1:], match.mBlockSizes[1:]):
                if last_s - s >= self.mIntronSize:
                    is_confirmed == False
                    # positions of exon boundaries on genome in alignment coordinates
                    pos1, pos2 = last_s - 1 - match.mSbjctFrom, s - match.mSbjctFrom
                    for mm in matches_with_introns:
                        if match.mQueryToken == mm.mQueryToken: continue
                        offset = match.mSbjctFrom - mm.mSbjctFrom
                        pp1 = pos1 + offset
                        pp2 = pos2 + offset
                        if pp1 >= 0 and pp2 >= 0 and \
                                match.mMapTarget2Query.mapRowToCol( pp1 ) >= 0 and \
                                match.mMapTarget2Query.mapRowToCol( pp2 ) >= 0 and \
                                match.mMapTarget2Query.mapRowToCol( pp2 ) - match.mMapTarget2Query.mapRowToCol( pp1 ) == 1:
                            is_confirmed = True
                            break
                    if is_confirmed:
                        new_matches_with_introns.append( match )
                        break

                new_matches_with_introns.append( match )
            
        return matches_without_introns + matches_with_introns

class FilterExonExtenders(Filter):
    """remove matches without introns that extend over an intron/exon boundary 
    created by another read.

    Introns need to be at least mMinIntronSize residues long.
    """
    mName = "filter_exon_extenders"

    # minimum intron size
    mMinIntronSize = 30

    # at least this number of residues are extended
    mMinLengthExonExtender = 10

    def __init__(self, *args, **kwargs ):
        Filter.__init__( self, *args, **kwargs )

        self.mOutFile.write( "tName\ttStart\ttEnd\tqName\trName\tintron\tnOverlap\n" )

    @E.benchmark
    def filter( self, sbjct_id, sbjct_start, sbjct_end, matches ):
        """filter matches.
        not fully implemented.
        """

        if self.options.loglevel >= 2:
            options.stdlog.write( "# started filter %s with %i matches\n" %\
                                      (self.mName, len(matches) ) )
            options.stdlog.flush()
        
        matches_without_introns, matches_with_introns = [], []

        for match in matches:
            if match.mSbjctNGapsBases < self.mMinIntronSize:
                matches_without_introns.append( match )
            else:
                matches_with_introns.append( match )
                    
        new_matches_without_introns = []
        for match_without_introns in matches_without_introns:
            for match_with_introns in matches_with_introns:
                d = alignlib.getAlignmentShortestDistance( \
                    match_without_introns.mMapTarget2Query, 
                    match_with_introns.mMapTarget2Query, 
                    alignlib.RR )

                if d > 0: continue

                do_delete = False

                for intron_start, intron_end in match_with_introns.iterator_introns():

                    if intron_end - intron_start < self.mMinIntronSize:
                        continue

                    # check overlap with this match
                    intron_map = alignlib.makeAlignmentBlocks()
                    intron_map.addDiagonal( intron_start - sbjct_start, 
                                            intron_end - sbjct_start, 
                                            0 )

                    d = alignlib.getAlignmentShortestDistance( \
                        match_without_introns.mMapTarget2Query, 
                        intron_map,
                        alignlib.RR )

                    if d > 0: continue
                    intron_sequence = self.mGenomeFasta.getSequence( sbjct_id, "+", intron_start, intron_end )
                    intron_type, prime5, prime3 = Genomics.GetIntronType( intron_sequence, both_strands = True )
                    
                    overlap = alignlib.getAlignmentOverlap( match_without_introns.mMapTarget2Query, 
                                                            intron_map,
                                                            alignlib.RR )

                    if intron_type != "unknown" and overlap > self.mMinLengthExonExtender:
                        
                        if self.options.loglevel >= 1:
                            self.options.stdlog.write( "# deleting read: %s: extends an exon into a %s intron in %s by %i residues\n" %\
                                                           (match_without_introns.mQueryId, intron_type, 
                                                            match_with_introns.mQueryId, overlap ) )
                            
                        do_delete = True
                        self.mOutFile.write( "%s\t%i\t%i\t%s\t%s\t%s\t%i\n" % 
                                             (sbjct_id, 
                                              sbjct_start, 
                                              sbjct_end, 
                                              match_without_introns.mQueryId,
                                              match_with_introns.mQueryId,
                                              intron_type,
                                              overlap) )
                        self.mOutFile.flush()
                        break

                if do_delete: 
                    break
            else:
                new_matches_without_introns.append( match_without_introns )

        return matches_with_introns + new_matches_without_introns

class FilterTranscriptMergers(Filter):
    """remove matches with introns that are not well supported.

    Introns need to be at least mMinIntronSize residues long.
    """
    
    mName = "filter_transcript_mergers"

    # minimum intron size to be deleted
    mMinIntronSize = 300

    # maximum minimum coverage in intron to be deleted
    mMaxCoverageIntron = 1.0

    # minimum avegare coverage in adjacent exons to consider intron for deletion
    mMinCoverageExons = 2.0

    def __init__(self, *args, **kwargs ):
        Filter.__init__( self, *args, **kwargs )

        self.mOutFile.write( "tName\ttStart\ttEnd\tqName\tiStart\tiEnd\tiLen\tcExons\tcIntrons\n" )
                             
    @E.benchmark
    def filter( self, sbjct_id, sbjct_start, sbjct_end, matches ):
        """filter matches.
        not fully implemented.
        """

        if self.options.loglevel >= 2:
            options.stdlog.write( "# started filter %s with %i matches\n" %\
                                      (self.mName, len(matches) ) )
            options.stdlog.flush()

        lgenome = sbjct_end - sbjct_start
        is_exon, is_intron, is_terminal = self.getReadCounts( matches, lgenome )

        new_matches = []

        for match in matches:

            keep = True

            exons = match.iterator_exons()
            last_start,last_end = exons.next()
            last_start -= sbjct_start
            last_end -= sbjct_start
            for start,end in exons:
                start -= sbjct_start
                end -= sbjct_start

                if start - last_end < self.mMinIntronSize: continue

                coverage_exons = float(numpy.sum( is_exon[last_start:last_end] ) + numpy.sum( is_exon[start:end] )) / \
                    ( last_end - last_start + end - start )

                v = [ is_intron[x] for x in filter( lambda x: is_exon[x] == 0, range(last_end,start)) ] 
                if len(v) == 0: continue
                coverage_intron = min ( v )
                
                lintron = start - last_end
                
                if coverage_intron <= self.mMaxCoverageIntron and coverage_exons >= self.mMinCoverageExons:
                    if options.loglevel >= 3:
                        options.stdlog.write( "# read: %s with intron %i-%i (l=%i): mean_cov_exons=%5.2f, min_cov_intron=%5.2f\n" %\
                                                  (match.mQueryId, 
                                                   sbjct_start + last_end, 
                                                   sbjct_start + start, 
                                                   lintron,
                                                   coverage_exons,
                                                   coverage_intron ) )

                    self.mOutFile.write( "%s\t%i\t%i\t%s\t%i\t%i\t%i\t%5.2f\t%5.2f\n" % 
                                         (sbjct_id, sbjct_start, sbjct_end, 
                                          match.mQueryId,
                                          sbjct_start + last_end, 
                                          sbjct_start + start, 
                                          lintron,
                                          coverage_exons,
                                          coverage_intron ) )

                    self.mOutFile.flush()
                    keep = False
                    break

                last_start, last_end = start, end

            if keep: new_matches.append( match )
            
        if options.loglevel >= 3:
            options.stdlog.write( "# %s - %s:%i..%i before=%i after=%i\n" %\
                                      (self.mName,
                                       sbjct_id, 
                                       sbjct_start,
                                       sbjct_end,
                                       len(matches),
                                       len(new_matches)))


        return new_matches

class FilterDuplicates(Filter):
    """remove duplicate reads with identical sequence from the list.
    """
    
    mName = "duplicates"

    # maximum difference in length
    mMinLengthDifference = 5

    mMinDistance = 10
    
    mMinAlignmentOverlap = 10

    mMinProbability = 0.05

    mMaxMatches = 10000

    def __init__(self, *args, **kwargs ):
        Filter.__init__( self, *args, **kwargs )

        self.mOutFile.write( "rep\tmem\tlRep\tlMem\tnoverlap\tdelta\tnreads\tPL\tPS\tP\n" )

        if options.loglevel >= 1:
            options.stdlog.write( "# %s: obtaining length distribution\n" % self.mName )
            options.stdlog.flush()

        ## estimate mean and stddev 
        lengths = self.mQueriesFasta.getLengths()
        summary = Stats.Summary( lengths )
        self.mLengthMean = summary.mMean
        self.mLengthMedian = summary.mMedian
        self.mLengthStd = summary.mSampleStd
                             
        if options.loglevel >= 1:
            options.stdlog.write( "# %s: estimated length distribution: n=%i, mean=%f, std=%f, median=%f\n" %\
                                      (self.mName,
                                       len(lengths),
                                       self.mLengthMean,
                                       self.mLengthStd,
                                       self.mLengthMedian) )
            options.stdlog.flush()

    @E.benchmark
    def filter( self, sbjct_id, sbjct_start, sbjct_end, matches ):
        """remove duplicate reads.

        Duplicate reads are mapped to the same position. The matches are sorted by
        increasing number of mismatches and gaps in query. With homopolymer runs
        in 454 data extra bases are often inserted.
        """

        if self.options.loglevel >= 2:
            options.stdlog.write( "# started filter %s with %i matches\n" %\
                                      (self.mName, len(matches) ) )
            options.stdlog.flush()

        new_matches = []
        keep = True
        eliminated = set()
        matches.sort( key=lambda x: x.mNMismatches + x.mQueryNGapsBases)

        lgenome = sbjct_end - sbjct_start
        is_exon, is_intron, is_terminal = self.getReadCounts( matches, lgenome )

        if len(matches) > self.mMaxMatches:
            return matches

        for x in range(len(matches)):
            if x in eliminated: continue
            mx = matches[x] 
            new_matches.append( mx )
            lx = mx.mQueryLength
            # seqx = self.mQueriesFasta.getSequence( mx.mQueryId )
            for y in range(x+1, len(matches)):
                if y in eliminated: continue
                my = matches[y]
                ly = my.mQueryLength

                # make sure they overlap
                overlap = alignlib.getAlignmentOverlap( \
                    mx.mMapTarget2Query, 
                    my.mMapTarget2Query, 
                    alignlib.RR )

                if overlap < self.mMinAlignmentOverlap: continue

                if mx.mSbjctFrom > my.mSbjctTo or mx.mSbjctTo < my.mSbjctFrom: continue
                if abs(mx.mSbjctFrom - my.mSbjctFrom) > self.mMinDistance or abs(mx.mSbjctTo - my.mSbjctTo) > self.mMinDistance: continue

                # compute 
                delta = abs( lx - ly )
                start = max(mx.mSbjctFrom,my.mSbjctFrom) - sbjct_start
                end = min(mx.mSbjctTo,my.mSbjctTo) - sbjct_start
                nreads = max( is_exon[start:end] )
                # P = 1.0 - ( 1.0 - float(delta+1) * min (lx, ly) / (lx * ly)) ** ((nreads * (nreads -1) ) / 2)
                # The factor 0.5 ensures that there is at least 1 residue in the interval if delta = 0.
                # are the brackets correct?
                PL = scipy.stats.norm.cdf( min(lx, lx + delta) + 0.5, loc = self.mLengthMean, scale = self.mLengthStd ) - \
                    scipy.stats.norm.cdf( max(0, lx - delta - 0.5), loc = self.mLengthMean, scale = self.mLengthStd )
                PS = float(delta+1) * min (lx, ly) / (lx * ly)
                # P = ( 1.0 - min(1.0, float(delta+1) * nreads * ( nreads-1 ) / 2 * min(lx,ly) / (lx * ly) ) ) * plx
                PSP = 1.0 - (1.0 - PS) ** ( nreads * (nreads-1) / 2.0 ) 
                PLP = 1.0 - (1.0 - PL) ** ( nreads * (nreads-1) / 2.0 )

                P = PLP * PSP
                # if abs( mx.mSbjctFrom - my.mSbjctFrom ) + abs(mx.mSbjctTo - my.mSbjctTo ) > self.mMinLengthDifference: continue
                # if abs(mx.mQueryLength - my.mQueryLength) > self.mMinLengthDifference: continue
                # seqy = self.mQueriesFasta.getSequence( my.mQueryId )
                if P <= self.mMinProbability:
                    self.mOutFile.write( "%s\t%s\t%i\t%i\t%i\t%i\t%i\t%f\t%f\t%f\n" % (mx.mQueryId, my.mQueryId, mx.mQueryLength, my.mQueryLength, overlap, delta, nreads, PL, PS, P ))
                    eliminated.add(y)
                    self.mOutFile.flush()

        if options.loglevel >= 3:
            options.stdlog.write( "# %s - %s:%i..%i before=%i after=%i\n" %\
                                      (self.mName,
                                       sbjct_id, 
                                       sbjct_start,
                                       sbjct_end,
                                       len(matches),
                                       len(new_matches)))


        return new_matches

class CoverageInfo: pass

def readMapId2Coverage( infile ):
    """read map of id to coverage information."""
    map_id2coverage = {}
    for line in infile:
        if line.startswith( "id" ): continue
        if line.startswith( "#" ): continue
        data = line[:-1].split("\t")
        x = CoverageInfo()
        id, x.mNMatches, x.mNCovered, x.mMean, x.mStddev = data[0], int(data[5]), int(data[6]), float(data[10]), float(data[12])
        map_id2coverage[id] = x

    return map_id2coverage

if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: psl2assembly.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"] )

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default]."  )

    parser.add_option("--input-filename-queries", dest="input_filename_queries", type="string",
                      help="fasta filename with queries [default=%default]."  )

    parser.add_option("--input-filename-coverage", dest="input_filename_coverage", type="string",
                      help="tabular table with coverage information [default=%default]."  )

    parser.add_option( "-p", "--output-filename-pattern", dest="output_filename_pattern", type="string" ,
                       help="OUTPUT filename with histogram information on aggregate coverages [%default].")

    parser.add_option( "--mali-output-format", dest="mali_output_format", type="choice" ,
                       choices=( "fasta", ),
                       help="output format to choose [default=%default].")

    parser.add_option( "--method", dest="methods", type="choice", action="append",
                       choices=( "region", "consensus", "pileup", "transcript", "coverage", "sbjct_coverage", "indels", "polyA", "locus" ),
                       help="methods to apply [%default].")

    parser.add_option( "-z", "--from-zipped", dest="from_zipped", action="store_true",
                       help="input is zipped.")

    parser.add_option( "--output-pileup", dest="output_pileup", action="store_true",
                       help="in the transcript modules, output the pileup alignment with the predicted transcript [default=%default].")

    parser.add_option( "--test", dest="test", type="int",
                       help="test - stop after # rows of parsing[%default]." )

    parser.add_option( "--staggered", dest="staggered_alignments", type="choice",
                       choices=("all", "none", "merge" ),
                       help="how to deal with staggered alignments[%default]." )

    parser.add_option( "--force-merge", dest="force_merge", type="int",
                       help="in case of staggered alignments, force merge if there are # alignments. This avoids costly computation of components in large sets. If 0, do not apply the threshold [%default]." )

    parser.add_option( "--filter", dest="filters", type="choice", action="append",
                       choices=("duplicates", "imperfect", "exon-extenders", "transcript-mergers"),
                       help="filters to apply [%default]." )

    parser.add_option( "--threshold-merge-distance", dest="threshold_merge_distance", type="int",
                       help="distance in nucleotides at which two adjacent reads shall be merged even if they are not overlapping [%default]." )

    parser.add_option( "--threshold-merge-overlap", dest="threshold_merge_overlap", type="int",
                       help="require reads to be overlapping by this amount to be joined into the same transcript [default=%default]." )

    parser.add_option( "--start-at", dest="start_at", type="int",
                       help="start numbering ids at number [default=%default]." )

    parser.set_defaults( input_filename_queries = None,
                         threshold_good_query_coverage = 90.0,
                         threshold_min_pid = 30.0,
                         output_filename_pattern = "%s",
                         mali_output_format = "fasta",
                         from_zipped = False,
                         genome_file = None,
                         test = None,
                         staggered_alignments = "none",
                         methods = [],
                         filters = [],
                         output_format = "%06i",
                         threshold_merge_distance = 1,
                         threshold_merge_overlap = 0,
                         output_pileup = False,
                         start_at = 1,
                         input_filename_coverage = None,
                         force_merge = None,
                         )

    (options, args) = E.Start( parser, add_pipe_options = True )

    if len(args) == 1:
        if options.from_zipped or args[0][-3:] == ".gz":
            import gzip
            infile = gzip.open( args[0], "r" )
        else:
            infile = open(args[0], "r" )
    else:
        infile = sys.stdin

    genome_fasta = IndexedFasta.IndexedFasta( options.genome_file )
    if options.input_filename_queries:
        queries_fasta =  IndexedFasta.IndexedFasta( options.input_filename_queries )
    else:
        queries_fasta = None

    if options.input_filename_coverage:
        map_id2coverage = readMapId2Coverage( open( options.input_filename_coverage, "r" ) )
        E.info("read coverage information for %i ids" % len(map_id2coverage) )
    else:
        map_id2coverage = None

    ################################################
    ################################################
    ################################################
    ## pick a processor
    ################################################        
    methods = []

    if len(options.methods) == 0:
        raise ValueError("please supply at least one method to apply.")

    for method in options.methods:
        if method == "region":
            methods.append( BuilderRegion( genome_fasta, queries_fasta, options ) )
        elif method == "pileup":
            methods.append( BuilderPileUp( genome_fasta, queries_fasta, options) )
        elif method == "consensus":
            methods.append( BuilderConsensus( genome_fasta, queries_fasta, options) )
        elif method == "locus":
            methods.append( BuilderTranscribedLocus( genome_fasta, queries_fasta, options) )
        elif method == "transcript":
            methods.append( BuilderTranscriptSpliced( genome_fasta, queries_fasta, options) )
        elif method == "coverage":
            methods.append( BuilderCoverage( genome_fasta, queries_fasta, options) )
        elif method == "sbjct_coverage":
            methods.append( BuilderSbjctCoverage( genome_fasta, queries_fasta, options) )
        elif method == "indels":
            methods.append( BuilderIndels( genome_fasta, queries_fasta, options) )
        elif method == "polyA":
            if not queries_fasta:
                raise ValueError( "need --input-filename-queries for polyA tail prediction." )
            methods.append( BuilderPolyA( genome_fasta, queries_fasta, options) )

    for method in methods:
        if map_id2coverage:
            method.setMapId2Coverage( map_id2coverage )
        method.printHeader()

    ################################################
    ################################################
    ################################################
    ## pick filters
    ## pre filters change components, post filters
    ## don't
    ################################################        
    pre_filters = []
    post_filters = []
    for f in options.filters:
        if f == "imperfect":
            post_filters.append( FilterImperfect( genome_fasta, queries_fasta, options ) )
        elif f == "exon-extenders":
            post_filters.append( FilterExonExtenders( genome_fasta, queries_fasta, options ) )
        elif f == "duplicates":
            pre_filters = [ FilterDuplicates( genome_fasta, queries_fasta, options ) ] + pre_filters
        elif f == "transcript-mergers":
            pre_filters.append( FilterTranscriptMergers( genome_fasta, queries_fasta, options ) )

    ################################################
    ################################################
    ################################################
    ## processing of a chunk (matches of same query)
    ################################################        
    ninput, noutput, nskipped, ncomponents = 0, 0, 0, 0

    output_id = options.start_at
    nskipped_components = 0
    n_pre_filtered_reads = 0
    n_post_filtered_reads = 0

    options.stdout.write( "%s\t%s\t%s\n" % ( "id","read", "strand" ))

    def processChunk( sbjct_id, sbjct_start, sbjct_end, matches ):
        global ninput, noutput, nskipped, ncomponents
        global nskipped_components
        global n_pre_filtered_reads
        global n_post_filtered_reads
        global output_id

        if len(matches) == 0:
            raise ValueError("no matches")

        ninput += 1

        if options.loglevel >= 2:
            options.stdlog.write( "## processing region: %s:%i..%i size=%i, matches=%i\n" %\
                                      (sbjct_id, sbjct_start, sbjct_end, sbjct_end - sbjct_start, len(matches)))
            options.stdlog.flush()

        # calculate overlapping components
        addAlignments( matches, shift = -sbjct_start, by_query = False )

        if options.loglevel >= 2:
            options.stdlog.write( "## region: %s:%i..%i: built alignments for %i matches\n" % \
                                      (sbjct_id, sbjct_start, sbjct_end, len(matches)) )

            options.stdlog.flush()


        if options.force_merge and len(matches) >= options.force_merge:
            options.stdlog.write( "## region: %s:%i..%i: forcing merge for %i matches\n" % \
                                      (sbjct_id, sbjct_start, sbjct_end, len(matches)) )
            options.stdlog.flush()
            components = ( range(0,len(matches), ), )
        else:
            components = Blat.getComponents( matches, 
                                            max_distance = options.threshold_merge_distance,
                                            min_overlap = options.threshold_merge_overlap )
            
            if options.loglevel >= 2:
                options.stdlog.write( "## region: %s:%i..%i: computed %i components\n" % \
                                          (sbjct_id, sbjct_start, sbjct_end, len(components)) )

                options.stdlog.flush()

        redo_components = False
        if pre_filters:
            new_matches = []
            for component in components:

                m = [ matches[ x ] for x in component ]

                before = len(m)
                for f in pre_filters:
                    m = f( sbjct_id, sbjct_start, sbjct_end, m )
                    if not m: break
                after = len(m)
                
                n_pre_filtered_reads += before - after
                
                if before - after > 0: 
                    redo_components = True
                
                new_matches += m

            if redo_components:
                matches = new_matches
                if options.force_merge and len(matches) >= options.force_merge:
                    options.stdlog.write( "## region: %s:%i..%i: forcing merge for %i matches\n" % \
                                              (sbjct_id, sbjct_start, sbjct_end, len(matches)) )
                    options.stdlog.flush()
                    components = ( range(0,len(matches), ), )
                else:
                    components = Blat.getComponents( matches, 
                                                     max_distance = options.threshold_merge_distance,
                                                     min_overlap = options.threshold_merge_overlap )

        if len(components) > 1:
            if options.staggered_alignments == "none":
                if options.loglevel >= 1:
                    options.stdlog.write( "## WARNING: more than one component (%i) in region %s:%i..%i (%s) - all are ignored\n" %\
                                              (len(components), sbjct_id, sbjct_start, sbjct_end, 
                                               str( [ len(x) for x in components ] )))
                nskipped += 1
                return
            elif options.staggered_alignments == "all":
                if options.loglevel >= 1:
                    options.stdlog.write( "## WARNING: more than one component (%i) in region %s:%i..%i (%s) - all are output\n" %\
                                              (len(components), sbjct_id, sbjct_start, sbjct_end, 
                                               str( [ len(x) for x in components ] )))
        
            elif options.staggered_alignments == "merge":
                if options.loglevel >= 1:
                    options.stdlog.write( "## WARNING: more than one component (%i) in region %s:%i..%i (%s) - all are merged\n" %\
                                              (len(components), sbjct_id, sbjct_start, sbjct_end, 
                                               str( [ len(x) for x in components ] )))
                    
                components = [ range(len(matches) ), ]

        ## applying filters
        for component in components:
            
            ncomponents += 1
            m = [ matches[ x ] for x in component ]

            before = len(m)
            for f in post_filters:
                m = f( sbjct_id, sbjct_start, sbjct_end, m )
                if not m: break
            after = len(m)

            n_post_filtered_reads += before - after

            if not m:
                if options.loglevel >= 1:
                    options.stdlog.write( "## region %s:%i..%i - component %i - all matches removed by filters\n" %\
                                              (sbjct_id, sbjct_start, sbjct_end, ncomponents ) ) 
                nskipped_components += 1
                continue

            for p in methods:
                p( output_id, sbjct_id, sbjct_start, sbjct_end, m )
                
            for match in m:
                options.stdout.write( "%s\t%s\t%s\n" %\
                                          (options.output_format % output_id,
                                           match.mQueryId,
                                           match.strand ) )
                
            options.stdout.flush()
            output_id += 1
            noutput += 1

    ################################################
    ################################################
    ################################################
    ## main loop: 
    ##
    ## collect overlapping entries on sbjct.
    ## Requirement: the input needs to be sorted
    ## by sbjct_id and sbjct_from
    ################################################        
    nfully_covered = None
    matches = []
    last_sbjct_id = None
    sbjct_start = None
    sbjct_end = None
    ninput_lines = 0

    skip = 0

    iterator = Blat.BlatIterator( sys.stdin )

    processed_contigs = set()

    while 1:

        match = iterator.next()
        
        if match == None: break
        
        ninput_lines += 1

        if options.test and ninput_lines > options.test:
            break
        
        if match.mSbjctId != last_sbjct_id or match.mSbjctFrom >= (sbjct_end + options.threshold_merge_distance):
            if last_sbjct_id:
                processChunk( last_sbjct_id, 
                              sbjct_start, sbjct_end,
                              matches )
            matches = []

            if last_sbjct_id != match.mSbjctId and match.mSbjctId in processed_contigs:
                raise ValueError("input not sorted correctly (contig,start): already encountered %s\n%s" % (match.mSbjctId, str(match)))
            last_sbjct_id = match.mSbjctId
            processed_contigs.add( last_sbjct_id )

            sbjct_start = match.mSbjctFrom
            sbjct_end = match.mSbjctTo
            
        if match.mSbjctFrom < sbjct_start:
            raise ValueError("input not sorted correctly (contig,start): %i < %i\n%s" % (match.mSbjctFrom, sbjct_start, str(match)))
        sbjct_end = max( match.mSbjctTo, sbjct_end )
        matches.append(match)

    processChunk( last_sbjct_id, sbjct_start, sbjct_end, matches )

    for p in methods:
        p.finish()

    if options.loglevel >= 1:
        options.stdlog.write("# clusters: ninput=%i, noutput=%i, ninput_lines=%i, ncomponents=%i\n" %\
                                 (ninput, noutput, ninput_lines, ncomponents ) )
        options.stdlog.write("# skipped: queries=%i, components=%i, pre_reads=%i, post_reads=%i\n" %\
                                 (nskipped, nskipped_components,
                                  n_pre_filtered_reads, n_post_filtered_reads))

    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
psl2fasta.py - output alignment pairs from psl formatted alignments
===================================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

analyze sequence pairs from a psl formatted table.

The sequences are assumed to be nucleotide sequences.

Usage
-----

Example::

   python psl2fasta.py --help

Type::

   python psl2fasta.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import tempfile
import subprocess
import optparse
import math

import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Blat as Blat
import CGAT.Genomics as Genomics

import psyco_full
import sys
import alignlib

##---------------------------------------------------------------------------------------------
def getAlignmentFull( m, q, t, options ):
    """print alignment with gaps in both query and target."""
    a = alignlib.AlignmentFormatExplicit( m, alignlib.makeSequence(q), alignlib.makeSequence(t) )
    return a.mRowAlignment, a.mColAlignment

##---------------------------------------------------------------------------------------------
    
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: psl2fasta.py 2781 2009-09-10 11:33:14Z andreas $",
                                    usage = globals()["__doc__"])

    parser.add_option("--filename-query", dest="filename_query", type="string",
                      help="fasta filename with queries."  )

    parser.add_option("--filename-target", dest="filename_target", type="string",
                      help="fasta filename with target."  )

    parser.add_option("-m", "--method", dest="method", type="choice",
                      choices=("full", "pileup-query", "pileup-target", "gapless" ),
                      help="method to use for constructing the alignment [%default]."  )

    parser.add_option("--forward-query", dest="forward_query", action="store_true",
                      help="reverse-complement sequences such that query is always on forward strand [%default]"  )

    parser.add_option("--target-prefix", dest="target_prefix", type="string",
                      help="prefix to use for target [%default]."  )

    parser.add_option("--query-prefix", dest="query_prefix", type="string",
                      help="prefix to use for query [%default]."  )

    parser.add_option("--id", dest="id", type="choice",
                      choices=("numeric", "query"),
                      help = "choose type of identifier to use [%default]" )

    parser.set_defaults(
        filename_query = None,
        filename_target = None,
        method = "full",
        output_format_id = "%06i",
        target_prefix = "",
        query_prefix = "",
        forward_query = False,
        )
    
    (options, args) = E.Start( parser )

    if options.filename_query:
        query = IndexedFasta.IndexedFasta( options.filename_query )

    if options.filename_target:
        target = IndexedFasta.IndexedFasta( options.filename_target )

    if options.method == "full":
        getAlignment = getAlignmentFull

    id = 0
    for match in Blat.iterator( options.stdin ):        
        if options.loglevel >= 2:
            options.stdout.write("# %s\n" % str(match))

        m = match.getMapQuery2Target()
        m.moveAlignment( -min(match.mQueryBlockStarts), -min(match.mSbjctBlockStarts) )
        q = query.getSequence( match.mQueryId, match.strand, match.mQueryFrom, match.mQueryTo )
        t = target.getSequence( match.mSbjctId, "+", match.mSbjctFrom, match.mSbjctTo )
        query_ali, sbjct_ali = getAlignment( m, q, t, options )

        if match.strand == "-" and options.forward_query:
            query_ali = Genomics.complement( query_ali )
            sbjct_ali = Genomics.complement( sbjct_ali )

        options.stdout.write(">%s%s:%s/%i-%i\n%s\n>%s%s:%s%s/%i-%i\n%s\n" % \
                                 (options.query_prefix, 
                                  options.output_format_id % id,
                                  match.mQueryId, match.mQueryFrom, match.mQueryTo,
                                  query_ali,
                                  options.target_prefix,
                                  options.output_format_id % id, 
                                  match.mSbjctId, match.strand, match.mSbjctFrom, match.mSbjctTo,
                                  sbjct_ali ) )
        id += 1

    E.Stop()
